# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Compressed array class
======================
"""
from __future__ import annotations

from typing import Any, Callable, Sequence, Union

import dask.array.backends
import dask.array.chunk_types
import dask.array.core
import dask.array.dispatch
import dask.base
import numcodecs.abc
import numpy
import numpy.lib.mixins
import zarr

from .type_hints import DType, NDArray, NDMaskedArray

#: Type of arrays returned when a compressed array is decompressed.
Array = Union[NDArray, NDMaskedArray]


class CompressedArray(numpy.lib.mixins.NDArrayOperatorsMixin):
    """Hold a compressed array and provide a numpy.ndarray interface to it.

    Each operation on the array uncompresses the data, performs the operation,
    an returns a numpy array. For the rechunk operation, the array is
    rechunked and the compressor is applied to the new chunks.

    Data is compressed using zarr and numcodecs.

    Args:
        array: A numpy.ndarray or a zarr.Array.
        *args: Arguments to pass to zarr.array.
        **kwargs: Keyword arguments to pass to zarr.array.
    """

    def __init__(
        self,
        array: CompressedArray | NDArray | zarr.Array,
        *args,
        **kwargs,
    ):
        self._fill_value = kwargs.get('fill_value', None)
        if isinstance(array, numpy.ndarray):
            self._array = zarr.array(array, *args, **kwargs)
        elif isinstance(array, zarr.Array):
            if args or kwargs:
                raise ValueError('args or kwargs are not allowed when '
                                 'array is a zarr.Array')
            self._array = array
        elif isinstance(array, CompressedArray):
            self._array = array._array
            self._fill_value = array._fill_value
        else:
            raise TypeError('array must be a numpy.ndarray or a zarr.Array')

    __array_priority__ = 0

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} ({self.shape}) {self.dtype}>'

    def _repr_html_(self) -> str:
        # pylint: disable=protected-access
        # Useless to rewrite the html representation of the array. Use the
        # zarr one.
        html_code: str = self._array.info._repr_html_()
        #: pylint: enable=protected-access
        return html_code.replace('zarr.core.Array', str(type(self)))

    def __getitem__(
        self,
        key: int | tuple[int | slice, ...],
    ) -> Array:
        """Retrieve data for an item or region of the array.

        Args:
            key: The key to retrieve.

        Returns:
            The data for the key.
        """
        values: Array = self._array[key]
        if self._fill_value is not None:
            values = numpy.ma.masked_equal(values, self._fill_value)
        return values

    def __setitem__(
        self,
        key: int | tuple[int | slice, ...],
        value: Array,
    ) -> None:
        """Set data for an item or region of the array.

        Args:
            key: The key to set.
            value: The value to set.
        """
        self._array[key] = value

    @property
    def ndim(self) -> int:
        """Number of dimensions."""
        return self._array.ndim

    @property
    def nchunks(self) -> int:
        """Number of chunks."""
        return self._array.nchunks

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of the array."""
        return self._array.shape

    @property
    def size(self) -> int:
        """The total number of elements in the array."""
        return self._array.size

    @property
    def dtype(self) -> DType:
        """The data type of the array."""
        return self._array.dtype

    @property
    def nbytes(self) -> int:
        """The total number of bytes used by the array."""
        return self._array.nbytes

    @property
    def compressor(self) -> numcodecs.abc.Codec | None:
        """The compressor used to compress the array."""
        return self._array.compressor

    @property
    def fill_value(self) -> Any:
        """The value used for uninitialized portions of the array."""
        return self._fill_value

    @property
    def filters(self) -> list[numcodecs.abc.Codec] | None:
        """One or more codecs used to transform data prior to compression."""
        return self._array.filters

    def reshape(self, *shape: int) -> Array:
        """Return a new array with the same data with a new shape.

        Args:
            *shape: The new shape.

        Returns:
            The reshaped array.
        """
        return self.__array__().reshape(*shape)

    @property
    def chunks(self) -> Sequence[Sequence[int]]:
        """A Tuple of integers describing the length of each dimension of a
        chunk of the array."""
        return self._array.chunks

    def rechunk(self, chunks: Sequence[Sequence[int]]) -> CompressedArray:
        """Rechunk the array.

        Args:
            chunks: The new chunks.

        Returns:
            The rechunked array.
        """
        return CompressedArray(
            zarr.array(self._array,
                       chunks=chunks,
                       compressor=self.compressor,
                       fill_value=self.fill_value,
                       filters=self.filters))

    def __array__(self, dtype=None) -> Array:
        """Return a numpy.ndarray of the array.

        Args:
            dtype: The data type of the array.

        Returns:
            The numpy.ndarray of the array or numpy.ma.masked_array if
            a fill_value is defined.
        """
        array: Array = self._array[...]
        if self._fill_value is not None:
            array = numpy.ma.masked_equal(array, self._fill_value)
        if dtype is not None:
            array = array.astype(dtype)
        return array

    def _cast(self, obj: Any) -> Any:
        """Cast an object to a value compatible with numpy.ndarray."""
        if isinstance(obj, CompressedArray):
            #: pylint: disable=protected-access
            # obj is a CompressedArray, we can access its _array attribute.
            return obj._array[...] if self._fill_value is None else \
                numpy.ma.masked_equal(obj._array[...], self._fill_value)
            #: pylint: enable=protected-access
        if isinstance(obj, numpy.ndarray):
            return obj
        if isinstance(obj, dask.array.core.Array):
            return obj.compute()
        if isinstance(obj, (float, int, bool)):
            return obj
        raise TypeError(f'Cannot cast {type(obj)} to array like object')

    def __array_function__(
        self,
        func: Callable,
        _types: Sequence[type],
        args,
        kwargs,
    ) -> Any:
        """Support numpy functions.

        Args:
            func: The numpy function.
            types: The types of the arguments.
            args: The arguments.
            kwargs: The keyword arguments.

        Returns:
            The result of the numpy function.
        """
        arrays: list[NDArray] = [
            item.__array__() if isinstance(item, CompressedArray) else item
            for item in args
        ]
        return func(*arrays, **kwargs)

    def __array_ufunc__(
        self,
        ufunc: Callable,
        method: str,
        *args,
        **kwargs,
    ) -> Any:
        """Call a numpy ufunc on the array.

        Args:
            ufunc: The ufunc to call.
            method: The method to call.
            *args: The arguments to pass to the ufunc.
            **kwargs: The keyword arguments to pass to the ufunc.

        Returns:
            The result of the ufunc.
        """
        inputs = [self._cast(arg) for arg in args]

        if method in ['__call__', 'reduce']:
            result = getattr(ufunc, method)(*inputs, **kwargs)
            if numpy.isscalar(result):
                return result
            return result

        return NotImplemented

    def astype(self, astype_dtype=None, **_kwargs) -> Array:
        """Return a copy of the array with the specified data type.

        Args:
            astype_dtype: The data type to cast to.

        Returns:
            The casted array.
        """
        return self.__array__(astype_dtype)


def dask_array_from_compressed_array(
    array: CompressedArray,
    name: str | None = None,
    inline_array=False,
    **kwargs,
) -> dask.array.core.Array:
    """Create a dask array from a compressed array.

    Args:
        array: The compressed array.

    Returns:
        The dask array.
    """
    chunks: Sequence[Sequence[int]] = array.chunks
    if name is None:
        name = 'from-compressed-array-' + dask.base.tokenize(
            array, chunks, **kwargs)
    return dask.array.core.from_array(
        array,
        chunks,  # type: ignore
        name=name,
        inline_array=inline_array)


@dask.array.dispatch.concatenate_lookup.register(CompressedArray)
def _concatenate_compressed_array(arrays, **kwargs) -> NDArray:
    dtype: DType | None = kwargs.get('dtype', None)
    arr: list[NDArray] = [item.__array__(dtype=dtype) for item in arrays]
    if any(tuple(item.fill_value is not None for item in arrays)):
        return numpy.ma.concatenate(arr, **kwargs)
    return numpy.concatenate(arr, **kwargs)


@dask.array.dispatch.numel_lookup.register(CompressedArray)
def _numel_compressed_array(array, **kwargs) -> numpy.float64 | NDArray:
    #: pylint: disable=protected-access
    # array is a CompressedArray, we can access its _array attribute.
    # We reuse the implementation of dask array numel. Useless to
    # reimplement it.
    arr: NDArray = array.__array__(dtype=kwargs.get('dtype', None))
    if array._fill_value is not None:
        return dask.array.backends._numel_masked(arr, **kwargs)
    return dask.array.backends._numel_ndarray(arr, **kwargs)
    #: pylint: enable=protected-access



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing interface for the variable module
=========================================
"""
import numpy

from ..abc import not_equal


def test_variable_not_equal() -> None:
    """Test if two values are different."""
    assert not_equal(1, 2) is True
    assert not_equal(1, 1) is False
    assert not_equal(1, '1') is True
    assert not_equal(1, numpy.nan) is True
    assert not_equal(numpy.nan, numpy.nan) is False
    assert not_equal(numpy.nan, 1) is True
    assert not_equal(numpy.datetime64('NaT'), numpy.datetime64('NaT')) is False
    assert not_equal(numpy.datetime64('NaT'), 1) is True



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Get software version information
================================
"""
__version__ = "2023.5.0"



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Handle merging of datasets of a partition.
==========================================
"""
from __future__ import annotations

from typing import Protocol
import random
import shutil

import fsspec
import fsspec.implementations.local
import zarr.storage

from .. import dataset, storage, sync
from .time_series import merge_time_series

__all__ = ('MergeCallable', 'perform', 'merge_time_series')

#: Character set used to create a temporary directory.
CHARACTERS = 'abcdefghijklmnopqrstuvwxyz0123456789_'


#: pylint: disable=too-few-public-methods,duplicate-code
class MergeCallable(Protocol):
    """Protocol to merge datasets stored in a partition.

    A merge callable is a function that accepts an existing dataset
    present in a partition, a new dataset to merge, the partitioning
    dimension and the axis to merge on. It returns the merged dataset.
    """

    def __call__(
        self,
        existing_ds: dataset.Dataset,
        inserted_ds: dataset.Dataset,
        axis: str,
        partitioning_dim: str,
    ) -> dataset.Dataset:  # pylint: disable=duplicate-code
        """Call the partition function.

        Args:
            existing_ds: The existing dataset.
            inserted_ds: The inserted dataset.
            axis: The axis to merge on.
            partitioning_dim: The partitioning dimension.

        Returns:
            The merged dataset.
        """
        # pylint: disable=unnecessary-ellipsis
        # Ellipsis is necessary to make the function signature match the
        # protocol.
        ...  # pragma: no cover
        # pylint: enable=unnecessary-ellipsis

    #: pylint: enable=too-few-public-methods,duplicate-code


def _rename(
    fs: fsspec.AbstractFileSystem,
    source: str,
    dest: str,
) -> None:
    """Rename a directory on a file system.

    Args:
        fs: The file system.
        source: The source directory.
        dest: The destination directory.
    """
    if isinstance(fs, fsspec.implementations.local.LocalFileSystem):
        # fspec implementation of the local file system, copy the source
        # directory to the destination directory and remove the source
        # directory. This is not efficient. So we use the shutil
        # implementation to rename the directory.
        shutil.rmtree(dest, ignore_errors=True)
        shutil.move(source, dest)
        return

    fs.rm(dest, recursive=True)
    fs.mv(source, dest, recursive=True)


def _update_fs(
    dirname: str,
    zds: dataset.Dataset,
    fs: fsspec.AbstractFileSystem,
    *,
    synchronizer: sync.Sync | None = None,
) -> None:
    """Updates a dataset stored in a partition.

    Args:
        dirname: The name of the partition.
        zds: The dataset to update.
        fs: The file system that the partition is stored on.
        synchronizer: The instance handling access to critical resources.
    """
    # Name of the temporary directory.
    temp: str = dirname + '.' + ''.join(
        random.choice(CHARACTERS) for _ in range(10))

    # Initializing Zarr group
    zarr.storage.init_group(store=fs.get_mapper(temp))

    # Writing new data.
    try:
        # The synchronization is done by the caller.
        storage.write_zarr_group(zds, temp, fs, synchronizer or sync.NoSync())
    except Exception:
        # The "write_zarr_group" method throws the exception if all scheduled
        # tasks are finished. So here we can delete the temporary directory.
        fs.rm(temp, recursive=True)
        raise

    # Rename the existing entry on the file system
    _rename(fs, temp, dirname)


def perform(
    ds_inserted: dataset.Dataset,
    dirname: str,
    axis: str,
    fs: fsspec.AbstractFileSystem,
    partitioning_dim: str,
    *,
    delayed: bool = True,
    merge_callable: MergeCallable | None,
    synchronizer: sync.Sync | None = None,
) -> None:
    """Merges a new dataset with an existing partition.

    Args:
        ds_inserted: The dataset to merge.
        dirname: The name of the partition.
        axis: The axis to merge on.
        fs: The file system on which the partition is stored.
        partitioning_dim: The partitioning dimension.
        delayed: If True, the existing dataset is loaded lazily. Defaults to
            True.
        merge_callable: The merge callable. If None, the inserted dataset
            overwrites the existing dataset stored in the partition.
            Defaults to None.
        synchronizer: The instance handling access to critical resources.
            Defaults to None.
    """
    zds: dataset.Dataset = merge_callable(
        storage.open_zarr_group(dirname, fs, delayed=delayed), ds_inserted,
        axis, partitioning_dim) if merge_callable is not None else ds_inserted
    _update_fs(dirname, zds, fs, synchronizer=synchronizer)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""Test the base class for indexing."""
from __future__ import annotations

from typing import Iterator
import pathlib

import fsspec
import numpy
import pyarrow
import pytest

from .. import abc
from ... import collection, convenience, dataset, partitioning
from ...partitioning.tests import data
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster
from ...tests.fs import local_fs
# pylint: enable=unused-import
from ...type_hints import NDArray


def split_half_orbit(
    cycle_number: numpy.ndarray,
    pass_number: numpy.ndarray,
) -> Iterator[tuple[int, int]]:
    """Calculate the indexes of the start and stop of each half-orbit.

    Args:
        pass_number: Pass numbers.

    Returns:
        Iterator of start and stop indexes.
    """
    assert pass_number.shape == cycle_number.shape
    pass_idx = numpy.where(numpy.roll(pass_number, 1) != pass_number)[0]
    cycle_idx = numpy.where(numpy.roll(cycle_number, 1) != cycle_number)[0]

    half_orbit = numpy.unique(
        numpy.concatenate(
            (pass_idx, cycle_idx, numpy.array([pass_number.size],
                                              dtype='int64'))))
    del pass_idx, cycle_idx

    yield from tuple(zip(half_orbit[:-1], half_orbit[1:]))


# pylint: disable=unused-argument,invalid-name
# The signature of the function must follow the signature of
# zcollection.PartitionCallable
def _half_orbit(
    zds: dataset.Dataset,
    *args,
    **kwargs,
) -> NDArray:
    """Return the indexes of the start and stop of each half-orbit.

    Args:
        zds: Datasets stored in a partition to be indexed.

    Returns:
        Dictionary of start and stop indexes for each half-orbit.
    """
    pass_number_varname = kwargs.pop('pass_number', 'pass_number')
    cycle_number_varname = kwargs.pop('cycle_number', 'cycle_number')
    pass_number = zds.variables[pass_number_varname].values
    cycle_number = zds.variables[cycle_number_varname].values

    generator = ((
        i0,
        i1,
        cycle_number[i0],
        pass_number[i0],
    ) for i0, i1 in split_half_orbit(cycle_number, pass_number))

    return numpy.fromiter(  # type: ignore
        generator, numpy.dtype(HalfOrbitIndexer.dtype()))


class HalfOrbitIndexer(abc.Indexer):
    """Index SWOT collection by half-orbit."""
    #: Column name of the cycle number.
    CYCLE_NUMBER = 'cycle_number'

    #: Column name of the pass number.
    PASS_NUMBER = 'pass_number'

    @classmethod
    def dtype(cls, /, **kwargs) -> list[tuple[str, str]]:
        """Return the columns of the index.

        Returns:
            A tuple of (name, type) pairs.
        """
        return super().dtype() + [
            (cls.CYCLE_NUMBER, 'uint16'),
            (cls.PASS_NUMBER, 'uint16'),
        ]

    @classmethod
    def create(
        cls,
        path: pathlib.Path | str,
        zds: collection.Collection,
        *,
        filesystem: fsspec.AbstractFileSystem | None = None,
        **kwargs,
    ) -> HalfOrbitIndexer:
        """Create a new index.

        Args:
            path: The path to the index.
            zds: The collection to be indexed.
            filesystem: The filesystem to use.

        Returns:
            The created index.
        """
        return super()._create(path,
                               zds,
                               meta={'attribute': b'value'},
                               filesystem=filesystem)  # type: ignore

    def update(
        self,
        zds: collection.Collection,
        *,
        partition_size: int | None = None,
        npartitions: int | None = None,
        **kwargs,
    ) -> None:
        """Update the index.

        Args:
            zds: New data stored in the collection to be indexed.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            cycle_number: The name of the cycle number variable stored in the
                collection. Defaults to "cycle_number".
            pass_number: The name of the pass number variable stored in the
                collection. Defaults to "pass_number".
        """
        super()._update(zds, _half_orbit, partition_size, npartitions,
                        **kwargs)


def test_indexer(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
        local_fs,  # pylint: disable=redefined-outer-name
):
    """Test the base class of the indexer."""
    ds = dataset.Dataset.from_xarray(data.create_test_sequence(5, 20, 10))

    zcollection = convenience.create_collection(
        'time',
        ds,
        partitioning.Date(('time', ), 'M'),
        partition_base_dir=str(local_fs.collection),
        filesystem=local_fs.fs)
    zcollection.insert(ds, merge_callable=collection.merging.merge_time_series)

    indexer = HalfOrbitIndexer.create(str(
        local_fs.collection.joinpath('index.parquet')),
                                      zcollection,
                                      filesystem=local_fs.fs)

    # Index not yet created
    with pytest.raises(ValueError):
        _ = indexer.table

    assert indexer.dtype() == [('start', 'int64'), ('stop', 'int64'),
                               ('cycle_number', 'uint16'),
                               ('pass_number', 'uint16')]
    indexer.update(zcollection)
    assert isinstance(indexer.table, pyarrow.Table)

    selection = zcollection.load(indexer=indexer.query({'cycle_number': 2}))
    assert selection is not None
    assert set(selection.variables['cycle_number'].values) == {2}

    with pytest.raises(ValueError):
        indexer.query({'cycle_number': 3}, logical_op='X')

    with pytest.raises(ValueError):
        indexer.query({'X': 3})

    # Updating the index should not change the indexer.
    indexer.update(zcollection)
    other = zcollection.load(indexer=indexer.query({'cycle_number': 2}))
    assert other is not None
    assert numpy.all(
        other['observation'].values == selection['observation'].values)

    selection = zcollection.load(
        indexer=indexer.query({'cycle_number': [2, 4]}))
    assert selection is not None
    assert set(selection.variables['cycle_number'].values) == {2, 4}

    selection = zcollection.load(indexer=indexer.query({
        'cycle_number': [2, 4],
        'pass_number': 1
    }))
    assert selection is not None
    assert set(selection.variables['cycle_number'].values) == {2, 4}
    assert set(selection.variables['pass_number'].values) == {1}

    selection = zcollection.load(indexer=indexer.query({
        'cycle_number': [2, 4],
        'pass_number': [1, 5]
    }))
    assert selection is not None
    assert set(selection.variables['cycle_number'].values) == {2, 4}
    assert set(selection.variables['pass_number'].values) == {1, 5}

    indexer = HalfOrbitIndexer.open(str(
        local_fs.collection.joinpath('index.parquet')),
                                    filesystem=local_fs.fs)
    assert indexer.meta == {'attribute': b'value'}
    selection = zcollection.load(indexer=indexer.query({
        'cycle_number': [2, 4],
        'pass_number': [1, 5]
    }))
    assert selection is not None
    assert set(selection.variables['cycle_number'].values) == {2, 4}
    assert set(selection.variables['pass_number'].values) == {1, 5}

    indices = tuple(
        indexer.query({'cycle_number': [2, 4]}, only_partition_keys=False))
    assert tuple(item[0] for item in indices[0][0]) == (
        'cycle_number',
        'pass_number',
        'year',
        'month',
    )

    indexer = HalfOrbitIndexer('', filesystem=fsspec.filesystem('memory'))
    assert indexer.query({'cycle_number': [2, 4]}) == tuple()



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test of the time periods.
=========================
"""
import datetime
import pickle

import numpy
import pytest

from ..period import Period, PeriodRelation


def _datetime64(year: int, month: int, day: int) -> numpy.datetime64:
    return numpy.datetime64(datetime.datetime(year, month, day))


def _datetime(value: int) -> numpy.datetime64:
    return numpy.datetime64(value, 'D')


def _timedelta(value: int) -> numpy.timedelta64:
    return numpy.timedelta64(value, 'D')


def _period(start: int, end: int, within: bool = False) -> Period:
    return Period(_datetime(start), _datetime(end), within=within)


def _period1(within=False) -> Period:
    return _period(1, 10, within=within)


def _period2() -> Period:
    return _period(5, 30)


def _period3() -> Period:
    return _period(35, 81)


def test_interface() -> None:
    """Test the interface of the Period class."""
    period1 = _period1()
    assert period1.begin == _datetime(1)
    assert period1.last == _datetime(9)
    assert period1.end() == _datetime(10)
    assert period1.length() == _timedelta(9)
    assert not period1.is_null()
    assert isinstance(str(period1), str)

    period1 = _period1(within=True)
    assert period1.begin == _datetime(1)
    assert period1.last == _datetime(10)
    assert period1.end() == _datetime(11)
    assert period1.length() == _timedelta(10)
    assert not period1.is_null()

    period2 = _period2()
    assert period2.begin == _datetime(5)
    assert period2.last == _datetime(29)
    assert period2.end() == _datetime(30)
    assert period2.length() == _timedelta(25)
    assert not period2.is_null()


def test_cmp() -> None:
    """Test the comparison operators."""
    period1 = _period1()
    period2 = _period2()
    period3 = _period3()

    assert period1 == _period1()
    assert period1 != period2
    assert period1 < period3

    with pytest.raises(TypeError):
        assert period1 < 1
    assert not period1 == 1  # pylint: disable=unneeded-not
    assert period1 != 1


def test_shift() -> None:
    """Test the shift method."""
    period1 = _period1()
    period1.shift(_timedelta(5))
    assert period1 == _period(6, 15)
    period1.shift(_timedelta(-15))
    assert period1 == _period(-9, 0)


def test_relation() -> None:
    """Test the relations between periods."""
    period1 = _period1()
    period2 = _period2()
    period3 = _period3()

    assert period2.contains(_datetime(20))
    assert not period2.contains(_datetime(2))

    assert period1.contains(_period(2, 8))
    assert not period1.contains(period3)

    assert period1.intersects(period2)
    assert period2.intersects(period1)

    assert period1.is_adjacent(_period(-5, 1))
    assert period1.is_adjacent(_period(10, 20))
    assert not period1.is_adjacent(period3)

    assert period1.is_before(_datetime(15))
    assert period3.is_after(_datetime(15))

    assert period1.intersection(period2) == _period(5, 10)
    assert period1.intersection(period3).is_null()

    assert period1.merge(period2) == _period(1, 30)
    assert period1.merge(period3).is_null()

    assert period3.span(period1) == _period(1, 81)

    period2 = Period(period1.begin - numpy.timedelta64(10, 's'),
                     period1.end() - numpy.timedelta64(1, 'D'))
    assert period1.intersection(period2) == Period(period1.begin,
                                                   period2.end())
    assert period1.merge(period2) == Period(period2.begin, period1.end())

    period2 = Period(period1.begin + numpy.timedelta64(10, 's'),
                     period1.end() + numpy.timedelta64(1, 'D'))
    assert period1.intersection(period2) == Period(period2.begin,
                                                   period1.end())

    period1 = _period(10, 20)
    period2 = _period(15, 18)
    assert period1.intersection(period2) == period2


def test_zero_length_period() -> None:
    """Test the zero length period."""
    zero_len = Period.from_duration(_datetime(3), _timedelta(0))
    assert _period(1, 1) == Period.from_duration(_datetime(1), _timedelta(0))
    assert _period(3, 3) == zero_len

    # zero_length period always returns false for is_before & is_after
    assert not zero_len.is_before(_datetime(5))
    assert not zero_len.is_after(_datetime(5))
    assert not zero_len.is_before(_datetime(-5))
    assert not zero_len.is_after(_datetime(-5))

    assert zero_len.is_null()
    assert not zero_len.contains(_datetime(20))
    # a null_period cannot contain any points
    assert not zero_len.contains(_datetime(3))
    assert not zero_len.contains(_period(5, 8))

    period1 = _period1()
    assert period1.contains(zero_len)
    assert zero_len.intersects(period1)
    assert period1.intersects(zero_len)
    assert zero_len.is_adjacent(_period(-10, 3))
    assert _period(-10, 3).is_adjacent(zero_len)
    assert zero_len.intersection(period1) == zero_len
    period2 = _period2()
    assert zero_len.span(period2) == _period(3, 30)


def test_invalid_period() -> None:
    """Test the invalid period."""
    null_per = _period(5, 1)

    assert not null_per.is_before(_datetime(7))
    assert not null_per.is_after(_datetime(7))
    assert not null_per.is_before(_datetime(-5))
    assert not null_per.is_after(_datetime(-5))

    assert null_per.is_null()
    assert not null_per.contains(_datetime(20))
    assert not null_per.contains(_datetime(3))
    assert not null_per.contains(_period(7, 9))
    period1 = _period1()
    assert period1.contains(null_per)
    assert null_per.intersects(period1)
    assert period1.intersects(null_per)
    assert null_per.is_adjacent(_period(-10, 5))
    assert null_per.is_adjacent(_period(1, 10))

    assert null_per.span(_period3()) == _period(5, 81)


def test_invalid() -> None:
    """Test the invalid periods."""
    period1 = _period(0, -2)
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(-3)
    assert period1.end() == _datetime(-2)
    assert period1.length() == _timedelta(-2)
    assert period1.is_null()

    period1 = _period(0, -1)
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(-2)
    assert period1.end() == _datetime(-1)
    assert period1.length() == _timedelta(-1)
    assert period1.is_null()

    period1 = _period(0, 0)
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(-1)
    assert period1.end() == _datetime(0)
    assert period1.length() == _timedelta(0)
    assert period1.is_null()

    period1 = _period(0, 1)
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(0)
    assert period1.end() == _datetime(1)
    assert period1.length() == _timedelta(1)
    assert not period1.is_null()

    period1 = _period(0, 2)
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(1)
    assert period1.end() == _datetime(2)
    assert period1.length() == _timedelta(2)
    assert not period1.is_null()

    period1 = Period.from_duration(_datetime(0), _timedelta(-1))
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(-2)
    assert period1.end() == _datetime(-1)
    assert period1.length() == _timedelta(-1)
    assert period1.is_null()

    period1 = Period.from_duration(_datetime(0), _timedelta(-2))
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(-3)
    assert period1.end() == _datetime(-2)
    assert period1.length() == _timedelta(-2)
    assert period1.is_null()

    period1 = Period.from_duration(_datetime(0), _timedelta(0))
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(-1)
    assert period1.end() == _datetime(0)
    assert period1.length() == _timedelta(0)
    assert period1.is_null()

    period1 = Period.from_duration(_datetime(0), _timedelta(1))
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(0)
    assert period1.end() == _datetime(1)
    assert period1.length() == _timedelta(1)
    assert not period1.is_null()

    period1 = Period.from_duration(_datetime(0), _timedelta(2))
    assert period1.begin == _datetime(0)
    assert period1.last == _datetime(1)
    assert period1.end() == _datetime(2)
    assert period1.length() == _timedelta(2)
    assert not period1.is_null()

    period1 = _period(1, 1)
    period2 = _period(1, 2)
    period3 = _period(1, 3)
    assert period1.length() == _timedelta(0)
    assert period2.length() == _timedelta(1)
    assert period3.length() == _timedelta(2)
    assert period1.is_null()
    assert not period2.is_null()

    period1 = _period(1, 2)
    period1.shift(_timedelta(1))
    period2 = _period(2, 3)
    assert period1 == period2

    period1 = Period.from_duration(_datetime(5), _timedelta(3))
    period2 = _period(3, 10)
    period1.expand(_timedelta(2))
    assert period1 == period2


def test_period_get_relation() -> None:
    """Test the get_relation method."""

    #          ##################
    #  #####
    start = _datetime64(2000, 6, 1)
    end = _datetime64(2000, 6, 30)
    period1 = Period(start, end, within=True)

    start = _datetime64(2000, 1, 5)
    end = _datetime64(2000, 1, 20)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.AFTER
    assert period1.get_relation(period2).is_after()

    #          ##################
    #     #####
    start = _datetime64(2000, 1, 5)
    end = _datetime64(2000, 6, 1)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.START_TOUCHING
    assert period1.get_relation(period2).is_after_overlapping()

    #          ##################
    #      #####
    start = _datetime64(2000, 1, 5)
    end = _datetime64(2000, 6, 2)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.START_INSIDE
    assert period1.get_relation(period2).is_after_overlapping()

    #          ##################
    #          ######################
    start = _datetime64(2000, 6, 1)
    end = _datetime64(2000, 7, 15)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(
        period2) == PeriodRelation.INSIDE_START_TOUCHING
    assert period1.get_relation(period2).is_before_overlapping()

    #          ##################
    #          #####
    start = _datetime64(2000, 6, 1)
    end = _datetime64(2000, 6, 15)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(
        period2) == PeriodRelation.ENCLOSING_START_TOUCHING
    assert period1.get_relation(period2).contains()

    #          ##################
    #                #####
    start = _datetime64(2000, 6, 10)
    end = _datetime64(2000, 6, 15)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.ENCLOSING
    assert period1.get_relation(period2).contains()

    #          ##################
    #                       #####
    start = _datetime64(2000, 6, 10)
    end = _datetime64(2000, 6, 30)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(
        period2) == PeriodRelation.ENCLOSING_END_TOUCHING
    assert period1.get_relation(period2).contains()

    #          ##################
    #          ##################
    start = _datetime64(2000, 6, 1)
    end = _datetime64(2000, 6, 30)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.EXACT_MATCH
    assert period1.get_relation(period2).contains()

    #          ##################
    #      ##########################
    start = _datetime64(2000, 5, 1)
    end = _datetime64(2000, 7, 15)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.INSIDE
    assert period1.get_relation(period2).is_inside()

    #          ##################
    #      ######################
    start = _datetime64(2000, 5, 1)
    end = _datetime64(2000, 6, 30)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.INSIDE_END_TOUCHING
    assert period1.get_relation(period2).is_after_overlapping()

    #          ##################
    #                         #####
    start = _datetime64(2000, 6, 15)
    end = _datetime64(2000, 7, 15)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.END_INSIDE
    assert period1.get_relation(period2).is_before_overlapping()

    #          ##################
    #                           #####
    start = _datetime64(2000, 6, 30)
    end = _datetime64(2000, 7, 15)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.END_TOUCHING
    assert period1.get_relation(period2).is_before_overlapping()

    #          ##################
    #                              #####
    start = _datetime64(2000, 7, 10)
    end = _datetime64(2000, 7, 15)
    period2 = Period(start, end, within=True)

    assert period1.get_relation(period2) == PeriodRelation.BEFORE
    assert period1.get_relation(period2).is_before()


def test_pickle() -> None:
    """Test pickling."""
    period = _period1()
    assert pickle.loads(pickle.dumps(period)) == period


def test_invalid_type() -> None:
    """Test invalid type."""
    with pytest.raises(ValueError):
        Period(numpy.int64(1), numpy.int64(2))  # type: ignore



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing datasets
================
"""
import pickle

import numpy
import pytest
import xarray
import zarr

from .. import dataset, meta
from ..variable import Variable
from ..variable.tests.data import array, delayed_array
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster

# pylint enable=unused-import


def create_delayed_dataset(factory) -> dataset.Dataset:
    """Create a test dataset."""
    return dataset.Dataset(attrs=(dataset.Attribute(name='attr', value=1), ),
                           variables=(factory(), factory('var2')))


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test dataset creation."""
    zds = create_delayed_dataset(factory)
    assert zds.dimensions == {'x': 5, 'y': 2}
    assert zds.attrs == (dataset.Attribute(name='attr', value=1), )
    assert isinstance(str(zds), str)
    assert isinstance(repr(zds), str)
    assert zds.nbytes == 5 * 2 * 8 * 2

    var1: Variable = factory()
    assert numpy.all(zds.variables['var1'].values == var1.values)
    assert zds.variables['var1'].metadata() == var1.metadata()
    assert numpy.all(var1.values == zds['var1'].values)
    assert var1.metadata() == zds['var1'].metadata()
    with pytest.raises(KeyError):
        print(zds['varX'])

    var2: Variable = factory('var2')
    assert numpy.all(zds.variables['var2'].values == var2.values)
    assert id(zds['var2']) == id(zds.variables['var2'])
    assert zds.variables['var2'].metadata() == var2.metadata()
    assert isinstance(zds.metadata(), meta.Dataset)

    other = zds.compute()
    assert isinstance(other, dataset.Dataset)

    ds_dict = zds.to_dict()
    assert isinstance(ds_dict, dict)
    assert isinstance(ds_dict['var1'], numpy.ndarray)
    assert isinstance(ds_dict['var2'], numpy.ndarray)

    ds_dict = zds.to_dict(variables=['var1'])
    assert isinstance(ds_dict, dict)
    assert isinstance(ds_dict['var1'], numpy.ndarray)
    assert 'var2' not in ds_dict

    zds = zds.compute()
    assert isinstance(zds, dataset.Dataset)
    assert isinstance(zds.var1, dataset.Array)
    assert isinstance(zds.var2, dataset.Array)


def test_dataset_dimensions_conflict(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test dataset creation with dimensions conflict."""
    with pytest.raises(ValueError):
        dataset.Dataset([
            dataset.DelayedArray(
                name='var1',
                data=numpy.arange(10, dtype='int64').reshape(5, 2),
                dimensions=('x', 'y'),
                attrs=(dataset.Attribute(name='attr', value=1), ),
                compressor=zarr.Blosc(cname='zstd', clevel=1),
                fill_value=0,
                filters=(zarr.Delta('int64',
                                    'int32'), zarr.Delta('int32', 'int32'))),
            dataset.DelayedArray(name='var2',
                                 data=numpy.arange(20, dtype='int64'),
                                 dimensions=('x'),
                                 attrs=(dataset.Attribute(name='attr',
                                                          value=1), ),
                                 compressor=zarr.Blosc(cname='zstd', clevel=1),
                                 fill_value=0,
                                 filters=(zarr.Delta('int64', 'int32'),
                                          zarr.Delta('int32', 'int32'))),
        ])


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_xarray(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test xarray creation."""
    zds = create_delayed_dataset(factory)
    xr1 = zds.to_xarray()
    assert isinstance(xr1, xarray.Dataset)

    xr2 = dataset.Dataset.from_xarray(xr1).to_xarray()
    assert xr1 == xr2


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_isel(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test dataset selection."""
    zds = create_delayed_dataset(factory)
    selected_zds = zds.isel(slices={'x': slice(0, 2)})
    assert selected_zds.dimensions == {'x': 2, 'y': 2}
    assert selected_zds.attrs == (dataset.Attribute(name='attr', value=1), )
    assert numpy.all(
        selected_zds.variables['var1'].values == numpy.arange(4).reshape(2, 2))
    selected_zds = zds.isel(slices={'y': slice(0, 1)})
    assert selected_zds.dimensions == {'x': 5, 'y': 1}
    assert numpy.all(selected_zds.variables['var1'].values == numpy.arange(
        0, 10, 2).reshape(5, 1))

    # Cannot slice on something which is not a dimension
    with pytest.raises(ValueError, match='invalid dimension'):
        zds.isel(slices={'z': slice(0, 1)})

    with pytest.raises(ValueError, match='invalid dimension'):
        zds.isel(slices={'var1': slice(0, 1)})


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_delete(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test dataset deletion."""
    zds = create_delayed_dataset(factory)

    other = zds.delete([1], 'y')
    assert numpy.all(
        other.variables['var1'].values == numpy.arange(0, 10, 2).reshape(5, 1))
    assert numpy.all(
        other.variables['var2'].values == numpy.arange(0, 10, 2).reshape(5, 1))

    other = zds.delete([0], 'x')
    assert numpy.all(other.variables['var1'].values == numpy.arange(
        10).reshape(5, 2)[1:, :])


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_concat(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test concatenation of datasets."""
    zds1 = create_delayed_dataset(factory)
    zds2 = create_delayed_dataset(factory)
    zds3 = zds1.concat(zds2, 'y')

    matrix = numpy.arange(10).reshape(5, 2)
    assert numpy.all(zds3.variables['var1'].values == numpy.concatenate(
        (matrix, matrix), axis=1))

    zds3 = zds1.concat(zds2, 'x')
    assert numpy.all(zds3.variables['var1'].values == numpy.concatenate(
        (matrix, matrix), axis=0))

    with pytest.raises(ValueError):
        zds1.concat([], 'z')


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_pickle(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test pickling of datasets."""
    zds = create_delayed_dataset(factory)
    other = pickle.loads(pickle.dumps(zds))
    assert zds.attrs == other.attrs
    assert list(zds.variables) == list(other.variables)
    for varname, variable in zds.variables.items():
        assert variable.attrs == other.variables[varname].attrs
        assert variable.compressor == other.variables[varname].compressor
        assert variable.dimensions == other.variables[varname].dimensions
        assert variable.dtype == other.variables[varname].dtype
        assert variable.fill_value == other.variables[varname].fill_value
        assert variable.filters == other.variables[varname].filters
        assert variable.name == other.variables[varname].name


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_add_variable(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test for adding a variable."""
    zds = create_delayed_dataset(factory)
    var = meta.Variable('var2',
                        numpy.int64,
                        dimensions=('x', 'y'),
                        attrs=(dataset.Attribute('attr', 1), ),
                        compressor=zarr.Blosc(),
                        fill_value=255,
                        filters=(zarr.Delta('int64', 'int64'), ))
    zds.add_variable(var)

    assert zds.variables['var2'].attrs == (dataset.Attribute(name='attr',
                                                             value=1), )
    assert zds.variables['var2'].dtype == numpy.int64
    assert zds.variables['var2'].dimensions == ('x', 'y')
    assert zds.variables['var2'].compressor == zarr.Blosc()
    assert zds.variables['var2'].filters == (zarr.Delta('int64', 'int64'), )
    assert zds.variables['var2'].fill_value == 255
    assert zds.variables['var2'].name == 'var2'
    assert zds.variables['var2'].shape == (5, 2)
    assert numpy.ma.allequal(
        zds.variables['var2'].values,
        numpy.ma.masked_equal(numpy.full((5, 2), 255, 'int64'), 255))

    other = zds.select_vars(['var1'])
    assert list(other.variables) == ['var1']

    data = numpy.ones((5, 2), 'int64')
    zds.drops_vars('var2')
    assert 'var2' not in zds.variables

    var = meta.Variable('var2',
                        numpy.int64,
                        dimensions=('x', 'y'),
                        attrs=(dataset.Attribute('attr', 1), ),
                        compressor=zarr.Blosc(),
                        fill_value=255,
                        filters=(zarr.Delta('int64', 'int64'), ))
    zds.add_variable(var, data)
    assert numpy.ma.allequal(
        zds.variables['var2'].values,
        numpy.ma.masked_equal(numpy.full((5, 2), 1, 'int64'), 255))

    data = numpy.ones((10, 2), 'int64')
    with pytest.raises(ValueError, match='Conflicting sizes'):
        zds.add_variable(var, data)

    var = meta.Variable('var2',
                        numpy.int64,
                        dimensions=('xx', 'y'),
                        attrs=(dataset.Attribute('attr', 1), ),
                        compressor=zarr.Blosc(),
                        fill_value=255,
                        filters=(zarr.Delta('int64', 'int64'), ))
    with pytest.raises(ValueError, match='has dimension'):
        zds.add_variable(var)


def test_empty_dataset(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test empty dataset."""
    zds = dataset.Dataset([], attrs=[])
    assert zds.attrs == ()
    assert not list(zds.variables)
    assert (str(zds)) == """<zcollection.dataset.Dataset>
  Dimensions: ()
Data variables:
    <empty>"""


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_rename(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test renaming of datasets."""
    zds = create_delayed_dataset(factory)
    zds.rename({'var1': 'var3', 'var2': 'var4'})
    assert zds.variables['var3'].name == 'var3'
    assert zds.variables['var4'].name == 'var4'
    assert 'var1' not in zds.variables
    assert 'var2' not in zds.variables

    with pytest.raises(ValueError):
        zds.rename({'var3': 'var4'})


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_persist(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test persisting of datasets."""
    zds1 = create_delayed_dataset(factory)
    zds1.persist()
    zds1.persist()
    assert zds1.variables['var1'].values is not None
    assert zds1.variables['var2'].values is not None

    zds2 = create_delayed_dataset(factory)
    zds2.persist(compress=True)
    zds2.persist(compress=True)
    assert zds2.variables['var1'].values is not None
    assert zds2.variables['var2'].values is not None

    assert numpy.all(
        zds1.variables['var1'].values == zds2.variables['var1'].values)
    assert numpy.all(
        zds1.variables['var2'].values == zds2.variables['var2'].values)


@pytest.mark.parametrize('factory', [array, delayed_array])
def test_dataset_rechunk(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test rechunking of datasets."""
    zds1 = create_delayed_dataset(factory)
    zds2 = create_delayed_dataset(factory)
    zds2 = zds2.rechunk()

    assert numpy.all(
        zds1.variables['var1'].values == zds2.variables['var1'].values)
    assert numpy.all(
        zds1.variables['var2'].values == zds2.variables['var2'].values)

    zds2 = zds1.rechunk().persist(compress=True)
    assert numpy.all(
        zds1.variables['var1'].values == zds2.variables['var1'].values)
    assert numpy.all(
        zds1.variables['var2'].values == zds2.variables['var2'].values)


def test_dataset_merge(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test merging of datasets."""
    template = dataset.Dataset(
        [
            dataset.DelayedArray('var1', numpy.empty(10), ('x', )),
            dataset.DelayedArray('var2', numpy.empty(10), ('x', )),
            dataset.DelayedArray('var3', numpy.empty((10, 10)), ('x', 'y')),
            dataset.DelayedArray('var4', numpy.empty((10, 10)), ('x', 'y')),
        ],
        attrs=[dataset.Attribute('attr1', 1),
               dataset.Attribute('attr2', 2)])
    zds2 = dataset.Dataset(
        [
            dataset.DelayedArray('var5', numpy.empty(10), ('x', )),
            dataset.DelayedArray('var6', numpy.empty(10), ('x', )),
            dataset.DelayedArray('var7', numpy.empty((10, 10)), ('x', 'y')),
            dataset.DelayedArray('var8', numpy.empty((10, 10)), ('x', 'y')),
        ],
        attrs=[dataset.Attribute('attr3', 3),
               dataset.Attribute('attr4', 4)])
    zds1 = pickle.loads(pickle.dumps(template))
    zds1.merge(zds2)
    assert list(zds1.variables) == [
        'var1', 'var2', 'var3', 'var4', 'var5', 'var6', 'var7', 'var8'
    ]
    assert zds1.dimensions == {'x': 10, 'y': 10}
    assert zds1.variables['var1'].shape == (10, )
    assert zds1.variables['var2'].shape == (10, )
    assert zds1.variables['var3'].shape == (10, 10)
    assert zds1.variables['var4'].shape == (10, 10)
    assert zds1.variables['var5'].shape == (10, )
    assert zds1.variables['var6'].shape == (10, )
    assert zds1.variables['var7'].shape == (10, 10)
    assert zds1.variables['var8'].shape == (10, 10)
    assert zds1.attrs == (dataset.Attribute('attr1',
                                            1), dataset.Attribute('attr2', 2),
                          dataset.Attribute('attr3',
                                            3), dataset.Attribute('attr4', 4))

    zds2 = dataset.Dataset(
        [
            dataset.DelayedArray('var5', numpy.empty(10), ('a', )),
            dataset.DelayedArray('var6', numpy.empty((10, 10)), ('a', 'b')),
        ],
        attrs=[dataset.Attribute('attr3', 3),
               dataset.Attribute('attr4', 4)])
    zds1 = pickle.loads(pickle.dumps(template))
    zds1.merge(zds2)
    assert list(
        zds1.variables) == ['var1', 'var2', 'var3', 'var4', 'var5', 'var6']
    assert zds1.dimensions == {'x': 10, 'y': 10, 'a': 10, 'b': 10}
    assert zds1.variables['var1'].shape == (10, )
    assert zds1.variables['var2'].shape == (10, )
    assert zds1.variables['var3'].shape == (10, 10)
    assert zds1.variables['var4'].shape == (10, 10)
    assert zds1.variables['var5'].shape == (10, )
    assert zds1.variables['var6'].shape == (10, 10)

    zds2 = dataset.Dataset(
        [
            dataset.DelayedArray('var5', numpy.empty(10), ('z', )),
        ],
        attrs=[dataset.Attribute('attr3', 3),
               dataset.Attribute('attr4', 4)])
    zds1 = pickle.loads(pickle.dumps(template))
    zds1.merge(zds2)
    assert list(zds1.variables) == ['var1', 'var2', 'var3', 'var4', 'var5']
    assert zds1.dimensions == {'x': 10, 'y': 10, 'z': 10}
    assert zds1.variables['var1'].shape == (10, )
    assert zds1.variables['var2'].shape == (10, )
    assert zds1.variables['var3'].shape == (10, 10)
    assert zds1.variables['var4'].shape == (10, 10)
    assert zds1.variables['var5'].shape == (10, )

    zds2 = dataset.Dataset(
        [
            dataset.DelayedArray('var1', numpy.empty(10), ('x', )),
        ],
        attrs=[dataset.Attribute('attr1', 1),
               dataset.Attribute('attr2', 2)])
    zds1 = pickle.loads(pickle.dumps(template))
    with pytest.raises(ValueError):
        zds1.merge(zds2)

    zds2 = dataset.Dataset(
        [
            dataset.DelayedArray('var5', numpy.empty(20), ('x', )),
        ],
        attrs=[dataset.Attribute('attr3', 3),
               dataset.Attribute('attr4', 4)])
    zds1 = pickle.loads(pickle.dumps(template))
    with pytest.raises(ValueError):
        zds1.merge(zds2)


def test_dataset_select_variables_by_dims(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test selecting variables by dimensions."""
    zds = dataset.Dataset([
        dataset.DelayedArray('var1', numpy.empty(10), ('x', )),
        dataset.DelayedArray('var2', numpy.empty(10), ('y', )),
        dataset.DelayedArray('var3', numpy.empty((10, 10)), ('x', 'y')),
        dataset.DelayedArray('var4', numpy.empty(20), ('a', )),
        dataset.DelayedArray('var5', numpy.empty(20), ('b', )),
        dataset.DelayedArray('var6', numpy.empty((20, 20)), ('a', 'b')),
        dataset.DelayedArray('var7', numpy.int64(1), ()),
    ],
                          attrs=[
                              dataset.Attribute('attr1', 1),
                          ])
    selected = zds.select_variables_by_dims(('x', ))
    assert list(selected.variables) == ['var1', 'var3']
    assert selected.dimensions == {'x': 10, 'y': 10}

    selected = zds.select_variables_by_dims(('x', 'y'))
    assert list(selected.variables) == ['var1', 'var2', 'var3']
    assert selected.dimensions == {'x': 10, 'y': 10}

    selected = zds.select_variables_by_dims(('x', 'y', 'z'))
    assert list(selected.variables) == ['var1', 'var2', 'var3']
    assert selected.dimensions == {'x': 10, 'y': 10}

    selected = zds.select_variables_by_dims(('a', ))
    assert list(selected.variables) == ['var4', 'var6']
    assert selected.dimensions == {'a': 20, 'b': 20}

    selected = zds.select_variables_by_dims(('a', 'b'))
    assert list(selected.variables) == ['var4', 'var5', 'var6']
    assert selected.dimensions == {'a': 20, 'b': 20}

    selected = zds.select_variables_by_dims(('a', 'b', 'c'))
    assert list(selected.variables) == ['var4', 'var5', 'var6']
    assert selected.dimensions == {'a': 20, 'b': 20}

    selected = zds.select_variables_by_dims(('x', 'a'))
    assert list(selected.variables) == ['var1', 'var3', 'var4', 'var6']
    assert selected.dimensions == {'x': 10, 'y': 10, 'a': 20, 'b': 20}

    selected = zds.select_variables_by_dims(tuple())
    assert list(selected.variables) == ['var7']
    assert selected.dimensions == {}

    selected = zds.select_variables_by_dims(tuple(), predicate=False)
    assert list(selected.variables) == [
        'var1', 'var2', 'var3', 'var4', 'var5', 'var6'
    ]
    assert selected.dimensions == {'x': 10, 'y': 10, 'a': 20, 'b': 20}

    selected = zds.select_variables_by_dims(('x', ), predicate=False)
    assert list(selected.variables) == ['var2', 'var4', 'var5', 'var6', 'var7']



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""Tests for :class:`zcollection.compressed_array.CompressedArray`."""
from typing import Any

import dask.array.core
import dask.array.creation
import dask.array.random
import dask.array.reductions
import dask.array.routines
import dask.array.ufunc
import dask.array.utils
import numpy
import pytest

from ..compressed_array import CompressedArray
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster

# pylint: enable=unused-import

# pylint: disable=unnecessary-lambda # We keep the lambdas for readability
#: Functions to test
functions = [
    lambda x: x,
    lambda x: dask.array.ufunc.expm1(x),
    lambda x: 2 * x,
    lambda x: x / 2,
    lambda x: x**2,
    lambda x: x + x,
    lambda x: x * x,
    lambda x: x[0],
    lambda x: x[:, 1],
    lambda x: x[:1, :, 1:3],
    lambda x: x.T,
    lambda x: dask.array.routines.transpose(x, (1, 2, 0)),
    lambda x: dask.array.reductions.nanmean(x),
    lambda x: dask.array.reductions.nanmean(x, axis=1),
    lambda x: dask.array.reductions.nanmax(x),
    lambda x: dask.array.reductions.nanmin(x),
    lambda x: dask.array.reductions.nanprod(x),
    lambda x: dask.array.reductions.nanstd(x),
    lambda x: dask.array.reductions.nanvar(x),
    lambda x: dask.array.reductions.nansum(x),
    lambda x: dask.array.reductions.median(x, axis=0),
    lambda x: dask.array.reductions.nanargmax(x),
    lambda x: dask.array.reductions.nanargmin(x),
    lambda x: dask.array.reductions.nancumprod(x, axis=0),
    lambda x: dask.array.reductions.nancumsum(x, axis=0),
    lambda x: x.sum(),
    lambda x: x.moment(order=0),
    lambda x: x.mean(),
    lambda x: x.mean(axis=1),
    lambda x: x.std(),
    lambda x: x.std(axis=1),
    lambda x: x.var(),
    lambda x: x.var(axis=1),
    lambda x: x.dot(numpy.arange(x.shape[-1])),
    lambda x: x.dot(numpy.eye(x.shape[-1])),
    lambda x: dask.array.routines.tensordot(
        x, numpy.ones(x.shape[:2]), axes=[(0, 1),
                                          (0, 1)]),  # type: ignore[arg-type]
    lambda x: x.sum(axis=0),
    lambda x: x.max(axis=0),
    lambda x: x.min(axis=0),
    lambda x: x.sum(axis=(1, 2)),
    lambda x: x.astype(numpy.complex128),
    lambda x: x.map_blocks(lambda x: x * 2),
    lambda x: x.map_overlap(
        lambda x: x * 2, depth=0, trim=True, boundary='none'),
    lambda x: x.map_overlap(
        lambda x: x * 2, depth=0, trim=False, boundary='none'),
    lambda x: x.round(1),
    lambda x: x.reshape((x.shape[0] * x.shape[1], x.shape[2])),
    lambda x: abs(x),
    lambda x: x > 0.5,
    lambda x: x.rechunk((4, 4, 4)),
    lambda x: x.rechunk((2, 2, 1)),
    lambda x: numpy.isneginf(x),
    lambda x: numpy.isposinf(x),
]
# pylint: enable=unnecessary-lambda


@pytest.mark.filterwarnings(
    'ignore:Casting complex values to real discards the imaginary part')
@pytest.mark.parametrize('func', functions)
def test_basic(
        func,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test basic functionality."""
    values: numpy.ndarray = numpy.random.random((2, 3, 4))
    arr: dask.array.core.Array = dask.array.core.from_array(
        CompressedArray(values), chunks='auto')
    compressed_array: numpy.ndarray = func(arr).compute()
    array: numpy.ndarray = func(dask.array.core.from_array(values)).compute()
    assert compressed_array.shape == array.shape
    assert numpy.allclose(compressed_array, array)


def test_metadata(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test metadata."""
    y: dask.array.core.Array = dask.array.random.random((10, 10),
                                                        chunks=(5, 5))
    z = CompressedArray(y.compute())
    y = y.map_blocks(CompressedArray)  # type: ignore[assignment]

    # pylint: disable=protected-access
    assert isinstance(y._meta, numpy.ndarray)
    assert isinstance((y + 1)._meta, numpy.ndarray)
    assert isinstance(y[:5, ::2]._meta, numpy.ndarray)
    assert isinstance(
        y.rechunk((2, 2))._meta,  # type: ignore[arg-type]
        numpy.ndarray)
    assert isinstance((y - z), numpy.ndarray)
    assert isinstance(y.persist()._meta, numpy.ndarray)
    # pylint: enable=protected-access


def test_from_delayed_meta(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test from_delayed with meta."""

    def f() -> CompressedArray:
        return CompressedArray(numpy.eye(3))

    d: Any = dask.delayed(f)()  # type: ignore
    x: dask.array.core.Array = dask.array.core.from_delayed(
        d, shape=(3, 3), meta=CompressedArray(numpy.eye(1)))
    assert numpy.all(x.compute() == f()[...])  # type: ignore


def test_from_array(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test from_array."""
    x = CompressedArray(numpy.eye(10))
    d: dask.array.core.Array = dask.array.core.from_array(
        x, chunks=(5, 5))  # type: ignore[arg-type]

    # pylint: disable=protected-access
    assert isinstance(d._meta, numpy.ndarray)
    # pylint: enable=protected-access
    assert isinstance(d.compute(), numpy.ndarray)
    assert numpy.allclose(d.compute(), x)


def test_map_blocks(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test map_blocks."""
    x: dask.array.core.Array = dask.array.creation.eye(
        10, chunks=5)  # type: ignore[arg-type]
    y: dask.array.core.Array = x.map_blocks(
        CompressedArray)  # type: ignore[arg-type]
    # pylint: disable=protected-access
    assert isinstance(y._meta, numpy.ndarray)
    # pylint: enable=protected-access
    assert numpy.allclose(y.compute(), x.compute())


def test_compressed_masked_array(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test CompressedMaskedArray."""
    x: dask.array.core.Array = dask.array.creation.eye(
        10, chunks=5)  # type: ignore[arg-type]
    y: dask.array.core.Array = x.map_blocks(
        CompressedArray, fill_value=0)  # type: ignore[arg-type]
    # assert isinstance(y._meta, CompressedArray)
    assert isinstance(y[...].compute(), numpy.ma.MaskedArray)
    assert isinstance(y.compute(), numpy.ma.MaskedArray)
    assert y.mean().compute() == 1
    assert y.min().compute() == 1
    assert y.max().compute() == 1
    assert y.sum().compute() == 10
    assert y.std().compute() == 0
    assert (y * 2).mean().compute() == 2



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
File system tools
=================
"""
from __future__ import annotations

from typing import Any, Iterator, Sequence
import os

import fsspec

#: Path separator
SEPARATOR = '/'


def join_path(*args: str) -> str:
    """Join path elements."""
    return SEPARATOR.join(args)


def normalize_path(fs: fsspec.AbstractFileSystem, path: str) -> str:
    """Normalize the path.

    Args:
        fs: file system object
        path: path to test

    Returns:
        Normalized path.
    """
    # pylint: disable=protected-access
    # There is no public method to perform this operation.
    path = fs._strip_protocol(path)  # type: ignore[return-value]
    # pylint: enable=protected-access
    if path == '':
        path = fs.sep
    if fs.protocol in ('file', 'memory'):
        return os.path.normpath(path)
    return path


def get_fs(
    filesystem: fsspec.AbstractFileSystem | str | None = None
) -> fsspec.AbstractFileSystem:
    """Return the file system object from the input.

    Args:
        filesystem: file system object or file system name

    Returns:
        File system object.

    Example:
        >>> from fsspec.implementations.local import LocalFileSystem
        >>> get_fs("hdfs")
        >>> get_fs(LocalFileSystem("/tmp/swot"))
    """
    filesystem = filesystem or 'file'
    return (fsspec.filesystem(filesystem)
            if isinstance(filesystem, str) else filesystem)


def fs_walk(
    fs: fsspec.AbstractFileSystem,
    path: str,
    sort: bool = False,
) -> Iterator[tuple[str, list[str], list[str]]]:
    """Return the list of files and directories in a directory.

    Args:
        fs: file system object
        path: path to the directory
        sort: if True, the list of files and directories is sorted
            alphabetically

    Returns:
        Iterator of (path, directories, files).
    """
    dirs: list[str]
    files: list[str]

    dirs, files = [], []
    try:
        listing: list[dict[str, Any]] = fs.ls(path, detail=True)
    except (FileNotFoundError, OSError):
        yield '', [], []
        return

    for info in listing:
        # each info name must be at least [path]/part , but here
        # we check also for names like [path]/part/
        pathname: str = info['name'].rstrip(SEPARATOR)
        name: str = pathname.rsplit(SEPARATOR, 1)[-1]
        if info['type'] == 'directory' and pathname != path:
            # do not include "self" path
            dirs.append(pathname)
        else:
            files.append(name)

    def sort_sequence(sequence: list[str]) -> list[str]:
        """Sort the sequence if the user wishes."""
        return list(sorted(sequence)) if sort else sequence

    dirs = sort_sequence(dirs)
    yield path.rstrip(SEPARATOR), dirs, sort_sequence(files)

    for item in dirs:
        yield from fs_walk(fs, item, sort=sort)


def copy_file(
    source: str,
    target: str,
    fs_source: fsspec.AbstractFileSystem,
    fs_target: fsspec.AbstractFileSystem,
) -> None:
    """Copy a file from one location to another.

    Args:
        source: The name of the source file.
        target: The name of the target file.
        fs_source: The file system that the source file is stored on.
        fs_target: The file system that the target file is stored on.
    """
    with fs_source.open(source, 'rb') as source_stream:
        with fs_target.open(target, 'wb') as target_stream:
            target_stream.write(source_stream.read())  # type: ignore[arg-type]


def copy_files(
    source: Sequence[str],
    target: str,
    fs_source: fsspec.AbstractFileSystem,
    fs_target: fsspec.AbstractFileSystem,
) -> None:
    """Copy a list of files from one location to another.

    Args:
        source: The names of the source files.
        target: The name of the target directory.
        fs_source: The file system that the source files are stored on.
        fs_target: The file system that the target directory is stored on.
    """
    tuple(
        map(
            lambda path: copy_file(path,
                                   join_path(target, os.path.basename(path)),
                                   fs_source, fs_target), source))


def copy_tree(
    source: str,
    target: str,
    fs_source: fsspec.AbstractFileSystem,
    fs_target: fsspec.AbstractFileSystem,
) -> None:
    """Copy a directory tree from one location to another.

    Args:
        source: The name of the source directory.
        target: The name of the target directory.
        fs_source: The file system that the source directory is stored on.
        fs_target: The file system that the target directory is stored on.

    Raises:
        ValueError: If the target already exists.
    """
    if fs_target.exists(target):
        raise ValueError(f'Target {target} already exists')
    fs_target.mkdir(target)
    for root, dirs, files in tuple(fs_walk(fs_source, source)):
        for name in files:
            source_path: str = join_path(root, name)
            copy_file(source_path,
                      join_path(target, os.path.relpath(source_path, source)),
                      fs_source, fs_target)
        for source_path in dirs:
            fs_target.mkdir(
                join_path(target, os.path.relpath(source_path, source)))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test of the collections
=======================
"""
from __future__ import annotations

import concurrent.futures
import datetime
import io

import dask.array.core
import dask.distributed
import fsspec
import numpy
import pytest
import zarr

from ... import (
    collection,
    convenience,
    dataset,
    merging,
    meta,
    partitioning,
    storage,
    sync,
    variable,
)
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster
from ...tests.data import (
    DELTA,
    END_DATE,
    FILE_SYSTEM_DATASET,
    START_DATE,
    create_test_collection,
    create_test_dataset,
    create_test_dataset_with_fillvalue,
)
from ...tests.fixture import dask_arrays, numpy_arrays
from ...tests.fs import local_fs, s3, s3_base, s3_fs

# pylint: disable=unused-import


@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
def test_collection_creation(
    fs,
    request,
) -> None:
    """Test the creation of a collection."""
    tested_fs = request.getfixturevalue(fs)
    zds = next(create_test_dataset(False))
    zcollection = collection.Collection(
        axis='time',
        ds=zds.metadata(),
        partition_handler=partitioning.Date(('time', ), 'D'),
        partition_base_dir=str(tested_fs.collection),
        filesystem=tested_fs.fs)
    assert isinstance(str(zcollection), str)
    assert zcollection.immutable is False

    serialized = collection.Collection.from_config(str(tested_fs.collection),
                                                   filesystem=tested_fs.fs)
    assert serialized.axis == zcollection.axis
    assert serialized.metadata == zcollection.metadata
    assert serialized.partition_properties == zcollection.partition_properties
    assert serialized.partitioning.get_config(
    ) == zcollection.partitioning.get_config()

    with pytest.raises(ValueError):
        collection.Collection.from_config(str(tested_fs.collection.parent))

    with pytest.raises(ValueError):
        collection.Collection('time_tai', zds.metadata(),
                              partitioning.Date(('time', ), 'D'),
                              str(tested_fs.collection))

    with pytest.raises(ValueError):
        collection.Collection('time', zds.metadata(),
                              partitioning.Date(('time_tai', ), 'D'),
                              str(tested_fs.collection))

    with pytest.raises(ValueError):
        collection.Collection(
            axis='time',
            ds=zds.metadata(),
            mode='X',  # type: ignore[arg-type]
            partition_handler=partitioning.Date(('time', ), 'D'),
            partition_base_dir=str(tested_fs.collection),
            filesystem=tested_fs.fs)


# pylint: disable=too-many-statements
@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_insert(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arrays_type,
    fs,
    request,
    tmpdir,
) -> None:
    """Test the insertion of a dataset."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    datasets = list(create_test_dataset(delayed=False))
    zcollection = collection.Collection('time',
                                        datasets[0].metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        str(tested_fs.collection),
                                        filesystem=tested_fs.fs,
                                        synchronizer=sync.ProcessSync(
                                            str(tmpdir / 'lock.lck')))

    indices = numpy.arange(0, len(datasets))
    numpy.random.shuffle(indices)
    for idx in indices:
        zcollection.insert(datasets[idx],
                           merge_callable=merging.merge_time_series)

    data = zcollection.load(delayed=delayed)
    assert data is not None
    values = data.variables['time'].values
    assert numpy.all(values == numpy.arange(START_DATE, END_DATE, DELTA))

    # Adding same datasets once more (should not change anything)
    for idx in indices[:5]:
        zcollection.insert(datasets[idx])

    assert list(zcollection.partitions()) == sorted(
        list(zcollection.partitions()))

    data = zcollection.load(delayed=delayed)
    assert data is not None
    values = data.variables['time'].values
    assert numpy.all(values == numpy.arange(START_DATE, END_DATE, DELTA))

    values = data.variables['var1'].values
    numpy.all(values == numpy.vstack((numpy.arange(values.shape[0]), ) *
                                     values.shape[1]).T)

    values = data.variables['var2'].values
    numpy.all(values == numpy.vstack((numpy.arange(values.shape[0]), ) *
                                     values.shape[1]).T)

    data = zcollection.load(delayed=delayed, filters='year == 2020')
    assert data is None

    data = zcollection.load(delayed=delayed, filters='year == 2000')
    assert data is not None
    assert data.variables['time'].shape[0] == 61

    data = zcollection.load(delayed=delayed,
                            filters='year == 2000 and month == 4')
    assert data is not None
    dates = data.variables['time'].values
    assert numpy.all(
        dates.astype('datetime64[M]') == numpy.datetime64('2000-04-01'))

    data = zcollection.load(
        delayed=delayed, filters='year == 2000 and month == 4 and day == 15')
    assert data is not None
    dates = data.variables['time'].values
    assert numpy.all(
        dates.astype('datetime64[D]') == numpy.datetime64('2000-04-15'))

    data = zcollection.load(
        delayed=delayed,
        filters='year == 2000 and month == 4 and day in range(5, 25)')
    assert data is not None
    data = zcollection.load(delayed=delayed,
                            filters=lambda keys: datetime.date(2000, 4, 5) <=
                            datetime.date(keys['year'], keys['month'], keys[
                                'day']) <= datetime.date(2000, 4, 24))
    assert data is not None
    dates = data.variables['time'].values.astype('datetime64[D]')
    assert dates.min() == numpy.datetime64('2000-04-06')
    assert dates.max() == numpy.datetime64('2000-04-24')

    for path, item in zcollection.iterate_on_records(relative=True):
        assert isinstance(path, str)
        assert isinstance(item, zarr.Group)

    zcollection = convenience.open_collection(str(tested_fs.collection),
                                              mode='r',
                                              filesystem=tested_fs.fs)
    zds = zcollection.load(delayed=delayed, selected_variables=['var1'])
    assert zds is not None
    assert 'var1' in zds.variables
    assert 'var2' not in zds.variables

    zds = zcollection.load(delayed=delayed, selected_variables=[])
    assert zds is not None
    assert len(zds.variables) == 0

    zds = zcollection.load(delayed=delayed, selected_variables=['varX'])
    assert zds is not None
    assert len(zds.variables) == 0

    # pylint: enable=too-many-statements


@pytest.mark.parametrize('fs,create_test_data', FILE_SYSTEM_DATASET)
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_update(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    create_test_data,
    request,
) -> None:
    """Test the update of a dataset."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    zds = next(create_test_data(delayed=delayed))
    zcollection = collection.Collection('time',
                                        zds.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        str(tested_fs.collection),
                                        filesystem=tested_fs.fs)
    zcollection.insert(zds)

    def update(zds: dataset.Dataset, shift: int = 3):
        """Update function used for this test."""
        return {'var2': zds.variables['var1'].values * -1 + shift}

    zcollection.update(update, delayed=delayed)  # type: ignore

    data = zcollection.load()
    assert data is not None
    assert numpy.allclose(data.variables['var2'].values,
                          data.variables['var1'].values * -1 + 3,
                          rtol=0)

    zcollection.update(
        update,  # type: ignore
        delayed=delayed,
        depth=1,
        shift=5)

    data = zcollection.load(delayed=delayed)
    assert data is not None
    assert numpy.allclose(data.variables['var2'].values,
                          data.variables['var1'].values * -1 + 5,
                          rtol=0)

    # Test case if the selected variables does not contains the variable
    # to update.
    zcollection.update(
        update,  # type: ignore
        delayed=delayed,
        selected_variables=['var1'],
        depth=1,
        shift=5)

    data = zcollection.load(delayed=delayed)
    assert data is not None
    assert numpy.allclose(data.variables['var2'].values,
                          data.variables['var1'].values * -1 + 5,
                          rtol=0)

    def update_with_info(zds: dataset.Dataset, partition_info=None, shift=3):
        """Update function used for this test."""
        assert partition_info is not None
        assert isinstance(partition_info, tuple)
        assert len(partition_info) == 2
        assert isinstance(partition_info[0], str)
        assert isinstance(partition_info[1], slice)
        assert partition_info[0] == 'num_lines'
        return {'var2': zds.variables['var1'].values * -1 + shift}

    zcollection.update(
        update_with_info,  # type: ignore
        delayed=delayed,
        depth=1,
        shift=10)

    data = zcollection.load(delayed=delayed)
    assert data is not None
    assert numpy.allclose(data.variables['var2'].values,
                          data.variables['var1'].values * -1 + 10,
                          rtol=0)

    def update_and_trim(zds: dataset.Dataset, partition_info=None):
        """Update function used for this test."""
        assert partition_info is not None
        assert partition_info[0] == 'num_lines'
        zds = zds.isel(dict((partition_info, )))
        return {'var2': zds.variables['var1'].values * -1}

    zcollection.update(
        update_and_trim,  # type: ignore
        delayed=delayed,
        trim=False,
        depth=1)

    data = zcollection.load(delayed=delayed)
    assert data is not None
    assert numpy.allclose(data.variables['var2'].values,
                          data.variables['var1'].values * -1,
                          rtol=0)

    def invalid_var_name(zds: dataset.Dataset):
        """Update function used to test if the user wants to update a non-
        existent variable name."""
        return {'var99': zds.variables['var1'].values * -1 + 3}

    with pytest.raises(ValueError):
        zcollection.update(invalid_var_name)  # type: ignore


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_drop_partitions(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
) -> None:
    """Test the dropping of a dataset."""
    tested_fs = request.getfixturevalue(arg)
    zcollection = create_test_collection(tested_fs, delayed=False)

    all_partitions = list(zcollection.partitions())
    assert 'month=01' in [
        item.split(zcollection.fs.sep)[-2] for item in all_partitions
    ]

    zcollection.drop_partitions(filters='year == 2000 and month==1')
    partitions = list(zcollection.partitions())
    assert 'month=01' not in [
        item.split(zcollection.fs.sep)[-2] for item in partitions
    ]

    npartitions = len(partitions)
    zcollection.drop_partitions(timedelta=datetime.timedelta(days=1))
    partitions = list(zcollection.partitions())
    assert len(partitions) == npartitions

    zcollection.drop_partitions(timedelta=datetime.timedelta(0))
    partitions = list(zcollection.partitions())
    assert len(partitions) == 0

    zcollection = convenience.open_collection(str(tested_fs.collection),
                                              mode='r',
                                              filesystem=tested_fs.fs)
    with pytest.raises(io.UnsupportedOperation):
        zcollection.drop_partitions()


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_drop_variable(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
) -> None:
    """Test the dropping of a variable."""
    tested_fs = request.getfixturevalue(arg)
    zcollection = create_test_collection(tested_fs, delayed=False)

    with pytest.raises(ValueError):
        zcollection.drop_variable('time')
    zcollection.drop_variable('var1')

    with pytest.raises(ValueError):
        zcollection.drop_variable('var1')

    zds = zcollection.load(delayed=False)
    assert zds is not None
    assert 'var1' not in zds.variables

    zcollection = convenience.open_collection(str(tested_fs.collection),
                                              mode='r',
                                              filesystem=tested_fs.fs)
    with pytest.raises(io.UnsupportedOperation):
        zcollection.drop_partitions()


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_add_variable(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
) -> None:
    """Test the adding of a variable."""
    tested_fs = request.getfixturevalue(arg)
    zcollection = create_test_collection(tested_fs, delayed=False)

    # Variable already exists
    new = meta.Variable(name='time',
                        dtype=numpy.dtype('float64'),
                        dimensions=('time', ))
    with pytest.raises(ValueError):
        zcollection.add_variable(new)

    # Variable doesn't use the partitioning dimension.
    new = meta.Variable(name='x',
                        dtype=numpy.dtype('float64'),
                        dimensions=('x', ))
    with pytest.raises(ValueError):
        zcollection.add_variable(new)

    # Variable doesn't use the dataset dimension.
    new = meta.Variable(name='x',
                        dtype=numpy.dtype('float64'),
                        dimensions=('time', 'x'))
    with pytest.raises(ValueError):
        zcollection.add_variable(new)

    new = meta.Variable(
        name='var3',
        dtype=numpy.dtype('int16'),
        dimensions=('num_lines', 'num_pixels'),
        fill_value=32267,
        attrs=(dataset.Attribute(name='attr', value=4), ),
    )
    zcollection.add_variable(new)

    assert new.name in zcollection.metadata.variables

    # Testing the configuration update by reopening the collection
    zcollection = collection.Collection.from_config(path=str(
        tested_fs.collection),
                                                    filesystem=tested_fs.fs)

    assert new.name in zcollection.metadata.variables

    zds = zcollection.load(delayed=False)
    assert zds is not None
    values = zds.variables['var3'].values
    assert isinstance(values, numpy.ma.MaskedArray)
    assert numpy.all(values.mask)  # type: ignore


@pytest.mark.parametrize('fs,create_test_data', FILE_SYSTEM_DATASET)
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_add_update(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    create_test_data,
    request,
) -> None:
    """Test the adding and updating of a dataset."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    zds = next(create_test_data(delayed=delayed))
    zcollection = collection.Collection('time',
                                        zds.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        str(tested_fs.collection),
                                        filesystem=tested_fs.fs)
    zcollection.insert(zds)

    new1 = meta.Variable(name='var3',
                         dtype=numpy.dtype('float64'),
                         dimensions=('num_lines', 'num_pixels'),
                         attrs=(dataset.Attribute(name='attr', value=1), ),
                         fill_value=1000000.5)

    new2 = meta.Variable(
        name='var4',
        dtype=numpy.dtype('int16'),
        dimensions=('num_lines', 'num_pixels'),
        fill_value=32267,
        attrs=(dataset.Attribute(name='attr', value=4), ),
    )
    zcollection.add_variable(new1)
    zcollection.add_variable(new2)

    data = zcollection.load(delayed=delayed)
    assert data is not None

    def update_1(zds, varname):
        """Update function used for this test."""
        return {varname: zds.variables['var1'].values * 201.5}

    def update_2(zds, varname):
        """Update function used for this test."""
        return {varname: zds.variables['var1'].values // 5}

    zcollection.update(update_1, new1.name, delayed=delayed)  # type: ignore
    zcollection.update(update_2, new2.name, delayed=delayed)  # type: ignore

    if delayed is False:
        # If the dataset is not delayed, we need to reload it.
        data = zcollection.load(delayed=False)
        assert data is not None

    assert numpy.allclose(data.variables[new1.name].values,
                          data.variables['var1'].values * 201.5,
                          rtol=0)
    assert numpy.allclose(data.variables[new2.name].values,
                          data.variables['var1'].values // 5,
                          rtol=0)


@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_fillvalue(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    request,
) -> None:
    """Test the management of masked values."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    zcollection = create_test_collection(tested_fs,
                                         delayed=delayed,
                                         with_fillvalue=True)

    # Load the dataset written with masked values in the collection and
    # compare it to the original dataset.
    data = zcollection.load(delayed=delayed)
    assert data is not None

    zds = next(create_test_dataset_with_fillvalue(delayed=delayed))

    values = data.variables['var1'].values
    assert isinstance(values, numpy.ma.MaskedArray)
    assert numpy.ma.allclose(zds.variables['var1'].values, values)

    values = data.variables['var2'].values
    assert isinstance(values, numpy.ma.MaskedArray)
    assert numpy.ma.allclose(zds.variables['var2'].values, values)


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_degraded_tests(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
) -> None:
    """Test the degraded functionality."""
    tested_fs = request.getfixturevalue(arg)
    zcollection = create_test_collection(tested_fs)

    fake_ds = next(create_test_dataset())
    fake_ds.variables['var3'] = fake_ds.variables['var1']
    fake_ds.variables['var3'].name = 'var3'

    with pytest.raises(ValueError):
        zcollection.insert(fake_ds)


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_insert_with_missing_variable(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
) -> None:
    """Test of the insertion of a dataset in which a variable is missing.

    This happens, for example, when a variable is not acquired, but
    created by an algorithm.
    """
    tested_fs = request.getfixturevalue(arg)
    zds = next(create_test_dataset_with_fillvalue()).to_xarray()
    zcollection = convenience.create_collection(
        axis='time',
        ds=zds,
        partition_handler=partitioning.Date(('time', ), 'M'),
        partition_base_dir=str(tested_fs.collection),
        filesystem=tested_fs.fs)
    zcollection.insert(zds, merge_callable=merging.merge_time_series)

    zds = next(create_test_dataset_with_fillvalue())
    zds.drops_vars('var1')
    zcollection.insert(zds)

    data = zcollection.load()
    assert data is not None
    assert numpy.ma.allequal(
        data.variables['var1'].values,
        numpy.ma.masked_equal(
            numpy.full(zds.variables['var1'].shape,
                       zds.variables['var1'].fill_value,
                       zds.variables['var1'].dtype),
            zds.variables['var1'].fill_value))


@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_insert_failed(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    request,
) -> None:
    """Test the insertion of a dataset in which the insertion failed."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    zds = next(create_test_dataset(delayed=delayed))
    zcollection = collection.Collection('time',
                                        zds.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        str(tested_fs.collection),
                                        filesystem=tested_fs.fs)
    partitions = list(zcollection.partitioning.split_dataset(zds, 'time'))

    # Create a file in the directory where the dataset should be written. This
    # should cause the insertion to fail.
    sep = zcollection.fs.sep
    one_directory = sep.join((zcollection.partition_properties.dir, ) +
                             partitions[0][0])
    zcollection.fs.makedirs(sep.join(one_directory.split(sep)[:-1]))
    zcollection.fs.touch(one_directory)

    with pytest.raises((OSError, ValueError)):
        zcollection.insert(zds)

    # Because the insert failed, the partition that was supposed to be created
    # was deleted.
    assert not zcollection.fs.exists(one_directory)
    zcollection.insert(zds)


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_insert_validation(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
) -> None:
    """Test the insertion of a dataset with metadata validation."""
    tested_fs = request.getfixturevalue(arg)
    zds = next(create_test_dataset_with_fillvalue())

    zcollection = convenience.create_collection(
        axis='time',
        ds=zds,
        partition_handler=partitioning.Date(('time', ), 'M'),
        partition_base_dir=str(tested_fs.collection),
        filesystem=tested_fs.fs)
    zcollection.insert(zds, merge_callable=merging.merge_time_series)

    zds = next(create_test_dataset_with_fillvalue())

    # Inserting a dataset containing valid attributes
    zcollection.insert(zds)

    # Inserting a dataset containing an invalid attributes
    zds = next(create_test_dataset_with_fillvalue())
    zds.attrs = (meta.Attribute('invalid', 1), )

    with pytest.raises(ValueError):
        zcollection.insert(zds, validate=True)

    # Inserting a dataset containing variables with invalid attributes
    zds = next(create_test_dataset_with_fillvalue())

    for var in zds.variables.values():
        var.attrs = (meta.Attribute('invalid', 1), )

    with pytest.raises(ValueError):
        zcollection.insert(zds, validate=True)


@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_map_partition(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    request,
) -> None:
    """Test the update of a dataset."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    zcollection = create_test_collection(tested_fs, delayed=delayed)

    result = zcollection.map(
        lambda x: x.variables['var1'].values * 2,  # type: ignore
        delayed=delayed)
    for item in result.compute():
        folder = zcollection.fs.sep.join(
            (zcollection.partition_properties.dir,
             zcollection.partitioning.join(item[0], zcollection.fs.sep)))
        zds = storage.open_zarr_group(folder, zcollection.fs, delayed=False)
        assert numpy.allclose(item[1], zds.variables['var1'].values * 2)


@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_indexer(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    request,
) -> None:
    """Test the update of a dataset."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    zcollection = create_test_collection(tested_fs, delayed=delayed)

    indexers = zcollection.map(
        lambda x: slice(0, x.dimensions['num_lines'])  # type: ignore
    ).compute()
    zds1 = zcollection.load(indexer=indexers, delayed=delayed)
    assert zds1 is not None
    zds2 = zcollection.load(delayed=delayed)
    assert zds2 is not None

    assert numpy.allclose(zds1.variables['var1'].values,
                          zds2.variables['var1'].values)


def test_variables() -> None:
    """Test the listing of the variables in a collection."""
    fs = fsspec.filesystem('memory')
    zds = next(create_test_dataset(delayed=False))
    zcollection = collection.Collection(axis='time',
                                        ds=zds.metadata(),
                                        partition_handler=partitioning.Date(
                                            ('time', ), 'D'),
                                        partition_base_dir='/',
                                        filesystem=fs)
    variables = zcollection.variables()
    assert isinstance(variables, tuple)
    assert len(variables) == 3
    assert variables[0].name == 'time'
    assert variables[1].name == 'var1'
    assert variables[2].name == 'var2'

    variables = zcollection.variables(('time', ))
    assert len(variables) == 1
    assert variables[0].name == 'time'


@pytest.mark.parametrize('fs', ['local_fs', 's3_fs'])
@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_map_overlap(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    fs,
    arrays_type,
    request,
) -> None:
    """Test the map overlap method."""
    tested_fs = request.getfixturevalue(fs)
    delayed = request.getfixturevalue(arrays_type)
    zcollection = create_test_collection(tested_fs, delayed=delayed)

    def func(zds: dataset.Dataset, partition_info: tuple[str, slice]):
        assert partition_info[0] == 'num_lines'
        var = zds.variables['var1']
        slices = [slice(None)] * len(var.dimensions)
        slices[var.dimensions.index(partition_info[0])] = partition_info[1]
        return var.values[tuple(slices)] * 2

    result = zcollection.map_overlap(
        func,  # type: ignore
        delayed=delayed,
        depth=1)

    for partition, data in result.compute():
        folder = zcollection.fs.sep.join(
            (zcollection.partition_properties.dir,
             zcollection.partitioning.join(partition, zcollection.fs.sep)))
        zds = storage.open_zarr_group(folder, zcollection.fs, delayed=False)
        assert numpy.allclose(data, zds.variables['var1'].values * 2)


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_insert_immutable(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    arg,
    request,
) -> None:
    """Test the insertion of a dataset with variables that are immutable
    relative to the partitioning."""
    tested_fs = request.getfixturevalue(arg)

    zds_reference = dataset.Dataset(
        [
            dataset.Array(
                'time',
                numpy.arange(numpy.datetime64('2000-01-01'),
                             numpy.datetime64('2000-01-30'),
                             numpy.timedelta64(1, 'D')),
                ('time', ),
            ),
            dataset.Array(
                'lon',
                numpy.arange(0, 360, 1),
                ('lon', ),
            ),
            dataset.Array(
                'lat',
                numpy.arange(-90, 90, 1),
                ('lat', ),
            ),
            dataset.Array(
                'grid',
                numpy.random.rand(
                    29,
                    360,
                    180,
                ),
                ('time', 'lon', 'lat'),
            ),
        ],
        attrs=(dataset.Attribute('history', 'Created for testing'), ),
    )
    zcollection = collection.Collection('time',
                                        zds_reference.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        str(tested_fs.collection),
                                        filesystem=tested_fs.fs)
    assert zcollection.immutable
    assert not tested_fs.fs.exists(zcollection._immutable)
    zcollection.insert(zds_reference)
    assert tested_fs.fs.exists(zcollection._immutable)

    zds = zcollection.load(delayed=False)
    assert zds is not None

    assert numpy.all(
        zds.variables['grid'].values == zds_reference.variables['grid'].values)
    assert numpy.all(
        zds.variables['time'].values == zds_reference.variables['time'].values)
    assert numpy.all(
        zds.variables['lon'].values == zds_reference.variables['lon'].values)
    assert numpy.all(
        zds.variables['lat'].values == zds_reference.variables['lat'].values)

    def update(zds: dataset.Dataset, varname: str) -> dict[str, numpy.ndarray]:
        """Update function used for this test."""
        return {varname: zds.variables['grid'].values * -1}

    zcollection.update(update, delayed=False, varname='grid')  # type: ignore
    zds = zcollection.load(delayed=False)
    assert zds is not None

    assert numpy.all(zds.variables['grid'].values ==
                     zds_reference.variables['grid'].values * -1)
    assert numpy.all(
        zds.variables['time'].values == zds.variables['time'].values)
    assert numpy.all(
        zds.variables['lon'].values == zds.variables['lon'].values)
    assert numpy.all(
        zds.variables['lat'].values == zds.variables['lat'].values)

    new_variable = meta.Variable(
        'new_var',
        numpy.float64,
        dimensions=('time', 'lon', 'lat'),
        attrs=(meta.Attribute('units', 'm'), ),
    )
    zcollection.add_variable(new_variable)
    zcollection.update(update, varname='new_var')  # type: ignore
    zds = zcollection.load()
    assert zds is not None
    assert numpy.all(zds.variables['new_var'].values ==
                     zds_reference.variables['grid'].values)

    new_variable = meta.Variable(
        'new_var2',
        numpy.float64,
        dimensions=('another_dim', ),
        attrs=(meta.Attribute('units', 'm'), ),
    )
    with pytest.raises(ValueError):
        zcollection.add_variable(new_variable)


@pytest.mark.parametrize('arg', ['local_fs', 's3_fs'])
def test_copy_collection(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
        arg,
        request,
        tmpdir) -> None:
    """Test the dropping of a dataset."""
    tested_fs = request.getfixturevalue(arg)
    zcollection = create_test_collection(tested_fs)

    target = str(tmpdir / 'copy')
    zcopy = zcollection.copy(target, filesystem=fsspec.filesystem('file'))

    ds_before_copy = zcollection.load()
    ds_after_copy = zcopy.load()

    assert ds_before_copy is not None
    assert ds_after_copy is not None

    assert numpy.all(ds_before_copy.variables['var1'].values ==
                     ds_after_copy.variables['var1'].values)
    assert numpy.all(ds_before_copy.variables['var2'].values ==
                     ds_after_copy.variables['var2'].values)


def _insert(zds: dataset.Dataset, base_dir: str, lock_file: str,
            scheduler_file: str) -> None:
    client = dask.distributed.Client(scheduler_file=scheduler_file)
    zcollection = collection.Collection.from_config(
        base_dir, mode='w', synchronizer=sync.ProcessSync(lock_file))
    zcollection.insert(zds, merge_callable=merging.merge_time_series)
    client.close()


# pylint: disable=too-many-statements
def test_concurrent_insert(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    tmpdir,
) -> None:
    """Test the insertion of a dataset."""
    fs = fsspec.filesystem('file')
    datasets = list(create_test_dataset(delayed=False))
    zds = datasets[0]
    lock_file = str(tmpdir / 'lock.lck')
    synchronizer = sync.ProcessSync(lock_file)
    base_dir = str(tmpdir / 'test')
    zcollection = collection.Collection('time',
                                        zds.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        base_dir,
                                        filesystem=fs,
                                        synchronizer=synchronizer)

    pool = concurrent.futures.ProcessPoolExecutor(max_workers=32)
    futures = []

    assert zcollection.is_locked() is False

    indices = list(numpy.arange(0, len(datasets)))
    numpy.random.shuffle(indices)
    futures = [
        pool.submit(_insert, datasets[ix], base_dir, lock_file,
                    dask_client.scheduler_file) for ix in indices
    ]

    def update(zds: dataset.Dataset, shift: int = 3):
        """Update function used for this test."""
        return {'var2': zds.variables['var1'].values * -1 + shift}

    launch_update = True
    for item in concurrent.futures.as_completed(futures):
        assert item.exception() is None
        if launch_update:
            zcollection.update(update)  # type: ignore
            launch_update = False

    data = zcollection.load()
    assert data is not None
    values = data.variables['time'].values

    assert numpy.all(values == numpy.arange(START_DATE, END_DATE, DELTA))


def test_partition_modified(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
        tmpdir):
    """Test the loading of a variable that has been modified since its
    creation."""
    fs = fsspec.filesystem('file')
    datasets = list(create_test_dataset())
    zds = datasets[0]
    base_dir = str(tmpdir / 'test')
    zcollection = collection.Collection('time',
                                        zds.metadata(),
                                        partitioning.Date(('time', ), 'D'),
                                        base_dir,
                                        filesystem=fs)
    zcollection.insert(zds)

    last_month = zcollection.load(filters=lambda keys: keys['year'] == 2000 and
                                  keys['month'] == 1 and keys['day'] == 16)
    assert last_month is not None

    zds = zcollection.load()
    assert zds is not None

    def new_shape(
        var: variable.Variable,
        selected_dim: str,
        new_size: int,
    ) -> tuple[int, ...]:
        """Compute the new shape of a variable."""
        return tuple(new_size if dim == selected_dim else size
                     for dim, size in zip(var.dimensions, var.shape))

    dim, size = 'num_lines', last_month.dimensions['num_lines'] * 25
    last_month = dataset.Dataset(
        [
            variable.DelayedArray(
                name,
                numpy.resize(var.array.compute(), new_shape(var, dim, size)),
                var.dimensions,
                attrs=var.attrs,
                compressor=var.compressor,
                fill_value=var.fill_value,
                filters=var.filters,
            ) for name, var in last_month.variables.items()
        ],
        attrs=zds.attrs,
    )
    zcollection.insert(last_month)

    with pytest.raises(RuntimeError, match='Try to re-load'):
        _ = zds['var2'].values

    with pytest.raises(RuntimeError, match='Try to re-load'):
        _ = zds['time'].values



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Mathematical functions.
=======================
"""
from typing import Iterable
import functools
import operator


def prod(iterable: Iterable) -> int:
    """Return the product of all elements in the given iterable.

    Args:
        iterable: An iterable containing numeric values.

    Returns:
        The product of all elements in the iterable. If the iterable is empty,
        returns 1.
    """
    return functools.reduce(operator.mul, iterable, 1)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Convenience functions
=====================
"""
from __future__ import annotations

import fsspec

from .. import collection, fs_utils, sync, view


def create_view(
    path: str,
    view_ref: view.ViewReference,
    *,
    filesystem: fsspec.AbstractFileSystem | str | None = None,
    filters: collection.PartitionFilter = None,
    synchronizer: sync.Sync | None = None,
) -> view.View:
    """Create a new view.

    Args:
        path: View storage directory.
        view_ref: Access properties for the reference view.
        filesystem: The file system used to access the view.
        filters: The filters used to select the partitions of the reference
            view. If not provided, all partitions are selected.
        synchronizer: The synchronizer used to synchronize the view.

    Example:
        >>> view_ref = ViewReference(
        ...     partition_base_dir="/data/mycollection")
        >>> view = create_view("/home/user/myview", view_ref)

    Returns:
        The created view.

    Raises:
        ValueError: If the path already exists.
    """
    filesystem = fs_utils.get_fs(filesystem)
    if filesystem.exists(path):
        raise ValueError(f'path {path!r} already exists.')
    return view.View(path,
                     view_ref,
                     ds=None,
                     filesystem=filesystem,
                     filters=filters,
                     synchronizer=synchronizer)


def open_view(
    path: str,
    *,
    filesystem: fsspec.AbstractFileSystem | str | None = None,
    synchronizer: sync.Sync | None = None,
) -> view.View:
    """Open an existing view.

    Args:
        path: View storage directory.
        filesystem: The file system used to access the view.
        synchronizer: The synchronizer used to synchronize the view.

    Returns:
        The opened view.

    Example:
        >>> view = open_view("/home/user/myview")
    """
    return view.View.from_config(path,
                                 filesystem=filesystem,
                                 synchronizer=synchronizer)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Make test data.
===============
"""
import numpy
import xarray

START_DATE = numpy.datetime64('2000-01-01', 'ns')
END_DATE = numpy.datetime64('2000-06-30', 'ns')
DELTA = numpy.timedelta64(72, 'h')


def create_test_sequence(repeatability, number_of_measures,
                         number_of_cycles) -> xarray.Dataset:
    """Creation of a data set for testing purposes."""
    pass_number = numpy.hstack([
        numpy.tile(ix + 1, number_of_measures) for j in range(number_of_cycles)
        for ix in range(repeatability)
    ])
    cycle_number = numpy.hstack([
        numpy.tile(ix + 1, repeatability * number_of_measures)
        for ix in range(number_of_cycles)
    ])
    delta = numpy.timedelta64(24 // repeatability // 2, 'h')
    time: numpy.ndarray = numpy.arange(START_DATE,
                                       START_DATE + len(cycle_number) * delta,
                                       delta)
    observation = numpy.random.rand(cycle_number.size)  # type: ignore
    return xarray.Dataset({
        'time':
        xarray.DataArray(
            time,
            dims=('num_lines', ),
        ),
        'cycle_number':
        xarray.DataArray(
            cycle_number,
            dims=('num_lines', ),
        ),
        'pass_number':
        xarray.DataArray(
            pass_number,
            dims=('num_lines', ),
        ),
        'observation':
        xarray.DataArray(
            observation,
            dims=('num_lines', ),
        ),
    })



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import importlib.metadata
import pathlib
import sys

HERE = pathlib.Path(__file__).absolute().parent

# Insert the project root dir as the first element in the PYTHONPATH.
sys.path.insert(0, str(HERE.parent.parent))

# -- Project information -----------------------------------------------------

project = 'zcollection'
copyright = '(2022, CNES/CLS)'
author = 'CLS'

# The full version, including alpha/beta/rc tags
try:
    release = importlib.metadata.version(project)
except importlib.metadata.PackageNotFoundError:
    release = '0.0.0'
version = '.'.join(release.split('.')[:2])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    'sphinx_inline_tabs',
    'sphinx_gallery.gen_gallery',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
]

autosummary_generate = True

autodoc_typehints = 'description'
autodoc_type_aliases = dict(
    ArrayLike='ArrayLike',
    DType='DType',
    DTypeLike='DTypeLike',
    Indexer='Indexer',
    NDArray='NDArray',
    NDMaskedArray='NDMaskedArray',
    PartitionCallback='PartitionCallback',
    QueryDict='QueryDict',
    Scalar='Scalar',
)

numpydoc_class_members_toctree = True
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'
html_title = 'ZCollection'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

intersphinx_mapping = {
    'dask': ('https://docs.dask.org/en/latest/', None),
    'fsspec': ('https://filesystem-spec.readthedocs.io/en/latest/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'python': ('https://docs.python.org/3/', None),
    'xarray': ('https://docs.xarray.dev/en/stable/', None),
    'zarr': ('https://zarr.readthedocs.io/en/stable', None),
}

# -- Extension configuration -------------------------------------------------
sphinx_gallery_conf = {
    'examples_dirs': [HERE.parent.parent.joinpath('examples')],
    'filename_pattern': r'[\\\/]ex_',
    'pypandoc': False,
}



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Fixtures for testing Dask clusters using the pytest.
====================================================
"""
from typing import Iterator
import contextlib
import logging
import weakref

import dask.config
import dask.distributed
import py
import pytest


@pytest.fixture()
def dask_cluster(
        pytestconfig,
        tmpdir_factory,
        scope='session',  # pylint: disable=unused-argument
) -> str:
    """Launch a Dask LocalCluster with a configurable number of workers."""
    n_workers: int | None
    threads_per_worker: int | None
    processes: bool

    try:
        n_workers = int(pytestconfig.getoption('n_workers'))
    except TypeError:
        n_workers = None

    try:
        threads_per_worker = int(pytestconfig.getoption('threads_per_worker'))
    except TypeError:
        threads_per_worker = None

    try:
        processes = int(pytestconfig.getoption('processes')) == 1
    except TypeError:
        processes = False

    tmpdir: py.path.local = tmpdir_factory.getbasetemp()
    scheduler_file: py.path.local = tmpdir / 'scheduler.json'
    if scheduler_file.exists():
        return str(scheduler_file)

    # Use the root path of the test session for the dask worker space
    dask_worker: py.path.local = tmpdir / 'dask_worker_space'
    dask.config.set(temporary_directory=str(dask_worker))

    logging.info('Dask local cluster starting')
    cluster = dask.distributed.LocalCluster(
        protocol='tcp://',
        n_workers=n_workers,
        threads_per_worker=threads_per_worker,
        processes=processes,
    )

    def teardown() -> None:
        """Stop the cluster and remove the scheduler file."""
        if scheduler_file.exists():
            scheduler_file.remove()

    weakref.finalize(cluster, teardown)

    # Make sure we can connect to the cluster.
    with dask.distributed.Client(cluster) as client:
        client.write_scheduler_file(scheduler_file)
        client.wait_for_workers(1)

    logging.info('Dask local cluster started')
    return str(scheduler_file)


@contextlib.contextmanager
def _scheduler_file(
        dask_cluster,  # pylint: disable=redefined-outer-name
) -> Iterator[str]:
    """Get the scheduler used to connect to the cluster."""
    yield dask_cluster


@pytest.fixture()
def dask_client(
    dask_cluster,  # pylint: disable=redefined-outer-name
) -> Iterator[dask.distributed.Client]:
    """Connect a Dask client to the cluster."""
    with _scheduler_file(dask_cluster) as scheduler_file:
        with dask.distributed.Client(scheduler_file=scheduler_file) as client:
            yield client



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test partitioning by date.
==========================
"""
from __future__ import annotations

import dataclasses
import pickle
import random
import string

import dask.array.core
import dask.local
import fsspec
import numpy
import pytest
import xarray

from .. import Date, get_codecs
from ... import dataset
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster
from ...type_hints import NDArray

# pylint: disable=disable=unused-argument

#: First date of the dataset to partition
START_DATE = numpy.datetime64('2000-01-06', 'ns')

#: Time delta between two partitions
TIME_DELTA = numpy.timedelta64(1, 'h')


@dataclasses.dataclass(frozen=True)
class PartitionTestData:
    """Test data for partitioning."""
    timedelta: numpy.timedelta64
    indices: slice
    resolution: str
    partitioning: Date
    dates: NDArray

    def check_partitioning(self, date: numpy.datetime64, zds: dataset.Dataset,
                           partition: tuple[str, ...]) -> None:
        """Check the partitioning of a dataset."""
        item = date.astype('datetime64[us]').item()
        assert partition == (
            f'year={item.year}',
            f'month={item.month:02d}',
            f'day={item.day:02d}',
            f'hour={item.hour:02d}',
        )[self.indices]

        folder = '/'.join(partition)
        fields = self.partitioning.parse(folder)
        parsed_date, = self.partitioning.encode(fields)
        assert parsed_date == numpy.datetime64(date).astype(
            f'datetime64[{self.resolution}]')

        expected_selection = self.dates[
            (self.dates >= parsed_date)
            & (self.dates < parsed_date + self.timedelta)]
        assert numpy.all(zds.variables['dates'].compute(
            scheduler=dask.local.get_sync) == expected_selection)

        assert fields == (
            ('year', item.year),
            ('month', item.month),
            ('day', item.day),
            ('hour', item.hour),
        )[self.indices]
        assert self.partitioning.join(fields, '/') == folder
        assert self.partitioning.join(
            self.partitioning.decode((parsed_date, )), '/') == folder


@pytest.mark.parametrize('delayed', [False, True])
def test_split_dataset(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    delayed: bool,
) -> None:
    """Test the split_dataset method."""
    for end_date, indices, resolution in [
        (
            numpy.datetime64('2001-12-31', 'Y'),
            slice(0, 1),
            'Y',
        ),
        (
            numpy.datetime64('2000-12-31', 'M'),
            slice(0, 2),
            'M',
        ),
        (
            numpy.datetime64('2000-12-31', 'D'),
            slice(0, 3),
            'D',
        ),
        (
            numpy.datetime64('2000-01-31', 'h'),
            slice(0, 4),
            'h',
        ),
    ]:

        # Time delta between two partitions
        timedelta = numpy.timedelta64(1, resolution)

        # Temporal axis to split
        dates: NDArray = numpy.arange(START_DATE, end_date, TIME_DELTA)

        # Measured data
        observation: NDArray = numpy.random.rand(dates.size)  # type: ignore

        # Create the dataset to split
        xds = xarray.Dataset({
            'dates':
            xarray.DataArray(dates, dims=('num_lines', )),
            'observation':
            xarray.DataArray(observation, dims=('num_lines', ))
        })

        partitioning = Date(('dates', ), resolution)
        assert len(partitioning) == len(range(indices.start, indices.stop))

        # Date of the current partition
        date = numpy.datetime64(START_DATE, resolution)

        # Build the test dataset
        zds = dataset.Dataset.from_xarray(xds)
        if not delayed:
            zds = zds.compute()

        checker = PartitionTestData(timedelta, indices, resolution,
                                    partitioning, dates)

        for partition, indexer in partitioning.split_dataset(zds, 'num_lines'):
            checker.check_partitioning(date, zds.isel(indexer), partition)
            date += timedelta


def test_construction() -> None:
    """Test the construction of the Date class."""
    partitioning = Date(('dates', ), 'D')
    assert partitioning.resolution == 'D'
    assert partitioning.variables == ('dates', )
    assert partitioning.dtype() == (('year', 'uint16'), ('month', 'uint8'),
                                    ('day', 'uint8'))
    assert len(partitioning) == 3
    assert partitioning.get_config() == {
        'id': 'Date',
        'resolution': 'D',
        'variables': ('dates', ),
    }

    with pytest.raises(ValueError):
        Date(('dates1', 'dates2'), 'D')

    with pytest.raises(ValueError):
        Date(('dates', ), 'W')


def test_config():
    """Test the configuration of the Date class."""
    partitioning = Date(('dates', ), 'D')
    assert partitioning.dtype() == (('year', 'uint16'), ('month', 'uint8'),
                                    ('day', 'uint8'))
    config = partitioning.get_config()
    partitioning = get_codecs(config)
    assert isinstance(partitioning, Date)


def test_pickle():
    """Test the pickling of the Date class."""
    partitioning = Date(('dates', ), 'D')
    other = pickle.loads(pickle.dumps(partitioning))
    assert isinstance(other, Date)
    assert other.resolution == 'D'
    assert other.variables == ('dates', )


@pytest.mark.parametrize('delayed', [False, True])
def test_no_monotonic(
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    delayed: bool,
):
    """Test that the Date partitioning raises an error if the temporal axis is
    not monotonic."""
    dates: numpy.ndarray = numpy.arange(numpy.datetime64('2000-01-01', 'h'),
                                        numpy.datetime64('2000-01-02', 'h'),
                                        numpy.timedelta64(1, 'm'))
    numpy.random.shuffle(dates)
    partitioning = Date(('dates', ), 'h')
    # pylint: disable=protected-access
    with pytest.raises(ValueError):
        arr = dask.array.core.from_array(dates) if delayed else dates
        list(partitioning._split({'dates': arr}))  # type: ignore[arg-type]
    # pylint: enable=protected-access


def test_values_must_be_datetime64(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test that the values must be datetime64."""
    dates = numpy.arange(numpy.datetime64('2000-01-01', 'h'),
                         numpy.datetime64('2000-01-02', 'h'),
                         numpy.timedelta64(1, 'm'))
    partitioning = Date(('dates', ), 'h')
    dates = dates.astype('int64')
    with pytest.raises(TypeError):
        # pylint: disable=protected-access
        list(
            partitioning._split({
                'dates':
                dask.array.core.from_array(dates)  # type: ignore[arg-type]
            }))
    # pylint: enable=protected-access


@pytest.mark.parametrize(
    'start, end, step, path_generator',
    [(('2000-01-01', 'D'), ('2000-02-01', 'D'), (1, 'D'), lambda item:
      (f'year={item.year}', f'month={item.month:02d}', f'day={item.day:02d}')),
     (('2000', 'Y'), ('2005', 'Y'), (1, 'Y'), lambda item:
      (f'year={item.year}', ))])
def test_listing_partition(start, end, step, path_generator):
    """Test the listing of the partitions."""
    fs = fsspec.filesystem('memory')
    variables = [
        'ancillary_surface_classification_flag',
        'correction_flag',
        'cross_track_angle',
        'cross_track_distance',
        'cycle_number',
        'dac',
        'depth_or_elevation',
        'distance_to_coast',
        'doppler_centroid',
        'dynamic_ice_flag',
        'geoid',
        'heading_to_coast',
        'height_cor_xover',
        'ice_conc',
        'internal_tide_hret',
        'internal_tide_sol2',
        'inv_bar_cor',
        'iono_cor_gim_ka',
        'latitude',
        'latitude_avg_ssh',
        'latitude_nadir',
        'load_tide_fes',
        'load_tide_got',
        'longitude',
        'longitude_avg_ssh',
        'longitude_nadir',
        'mean_dynamic_topography',
        'mean_dynamic_topography_uncert',
        'mean_sea_surface_cnescls',
        'mean_sea_surface_cnescls_uncert',
        'mean_sea_surface_dtu',
        'mean_sea_surface_dtu_uncert',
        'mean_wave_direction',
        'mean_wave_period_t02',
        'model_dry_tropo_cor',
        'model_wet_tropo_cor',
        'num_pt_avg',
        'obp_ref_surface',
        'ocean_tide_eq',
        'ocean_tide_fes',
        'ocean_tide_got',
        'ocean_tide_non_eq',
        'orbit_alt_rate',
        'orbit_qual',
        'pass_number',
        'phase_bias_ref_surface',
        'polarization_karin',
        'pole_tide',
        'rad_cloud_liquid_water',
        'rad_surface_type_flag',
        'rad_tmb_187',
        'rad_tmb_238',
        'rad_tmb_340',
        'rad_water_vapor',
        'rad_wet_tropo_cor',
        'rain_flag',
        'rain_rate',
        'sc_altitude',
        'sc_pitch',
        'sc_roll',
        'sc_yaw',
        'sea_state_bias_cor',
        'sea_state_bias_cor_2',
        'sig0_cor_atmos_model',
        'sig0_cor_atmos_rad',
        'sig0_karin',
        'sig0_karin_2',
        'sig0_karin_qual',
        'sig0_karin_uncert',
        'simulated_error_baseline_dilation',
        'simulated_error_karin',
        'simulated_error_orbital',
        'simulated_error_phase',
        'simulated_error_roll',
        'simulated_error_timing',
        'simulated_error_troposphere',
        'simulated_true_ssh_karin',
        'solid_earth_tide',
        'ssha_karin',
        'ssha_karin_2',
        'ssha_karin_qual',
        'ssh_karin',
        'ssh_karin_2',
        'ssh_karin_uncert',
        'swh_karin',
        'swh_karin_qual',
        'swh_karin_uncert',
        'swh_model',
        'swh_sea_state_bias',
        'time',
        'time_tai',
        'velocity_heading',
        'wind_speed_karin',
        'wind_speed_karin_2',
        'wind_speed_model_u',
        'wind_speed_model_v',
        'wind_speed_rad',
        'x_factor',
    ]
    root = '/' + ''.join(random.choices(string.ascii_letters, k=10))
    fs.mkdir(root)
    fs.open(fs.sep.join((root, '.zcollection')), 'w').close()

    partitioning = Date(('dates', ), step[1])

    start = numpy.datetime64(*start)
    end = numpy.datetime64(*end)
    step = numpy.timedelta64(*step)

    expected = []
    for date in numpy.arange(start, end, step):
        item = date.item()
        partition = fs.sep.join((root, *path_generator(item)))
        expected.append(partition)
        fs.mkdirs(partition)

        _ = {fs.mkdirs(fs.sep.join((partition, item))) for item in variables}
        _ = {
            fs.open(fs.sep.join((partition, item)), 'w').close()
            for item in ['.zattrs', '.zgroup', '.zmetadata']
        }

    assert expected == list(partitioning.list_partitions(fs, root))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Implementation details.
=======================
"""
from __future__ import annotations

from typing import Callable, Iterable, Iterator, Sequence, Tuple
import base64
import dataclasses
import hashlib
import io
import pathlib
import pickle
import warnings

import dask.array.core
import dask.bag.core
import dask.distributed
import fsspec
import numpy
import zarr

from .. import collection, dataset, meta
from ..collection.detail import _update_with_overlap
from ..fs_utils import get_fs, join_path
from ..storage import (
    DIMENSIONS,
    open_zarr_array,
    open_zarr_group,
    update_zarr_array,
    write_zattrs,
)
from ..type_hints import ArrayLike, NDArray

#: Type of the function used to update a view.
ViewUpdateCallable = \
    Callable[[Iterable[Tuple[dataset.Dataset, str]], str], None]

#: Name of the file that contains the checksum of the view.
CHECKSUM_FILE = '.checksum'

#: Name of the attribute that contains the checksum of the view.
CHECKSUM_ATTR = 'checksum'


@dataclasses.dataclass(frozen=True)
class ViewReference:
    """Properties of the collection used as reference by a view.

    Args:
        path: Path to the collection.
        filesystem: The file system used to access the reference collection.
    """
    #: Path to the collection.
    path: str
    #: The file system used to access the reference collection.
    filesystem: fsspec.AbstractFileSystem = get_fs('file')


@dataclasses.dataclass(frozen=True)
class AxisReference:
    """Properties of the axis used as reference by a view.

    Args:
        array: Axis values.
        dimension: Name of the dimension.
        checksum: Checksum of the axis values.
    """
    #: Axis values.
    array: NDArray
    #: Name of the dimension.
    dimension: str
    #: Checksum of the axis values.
    checksum: str


def _create_zarr_array(args: tuple[str, zarr.Group],
                       *,
                       base_dir: str,
                       fs: fsspec.AbstractFileSystem,
                       template: str,
                       variable: meta.Variable,
                       invalidate_cache: bool = True) -> None:
    """Create a Zarr array, with fill_value being used as the default value for
    uninitialized portions of the array.

    Args:
        args: tuple of (path, zarr.Group).
        base_dir: Base directory for the Zarr array.
        fs: The filesystem used to create the Zarr array.
        template: The variable's name is used as a template for the Zarr array
            to determine the shape of the new variable.
        variable: The properties of the variable to create.
        invalidate_cache: If True, invalidate the cache of the directory
            containing the Zarr array.
    """
    partition: str
    group: zarr.Group

    partition, group = args
    data: dask.array.core.Array = dask.array.core.from_zarr(group[template])

    dirname: str = join_path(base_dir, partition)
    mapper: fsspec.FSMap = fs.get_mapper(join_path(dirname, variable.name))
    zarr.full(data.shape,
              chunks=data.chunksize,
              dtype=variable.dtype,
              compressor=variable.compressor,
              fill_value=variable.fill_value,
              store=mapper,
              overwrite=True,
              filters=variable.filters)
    write_zattrs(dirname, variable, fs)
    if invalidate_cache:
        fs.invalidate_cache(dirname)


def _drop_zarr_zarr(partition: str,
                    fs: fsspec.AbstractFileSystem,
                    variable: str,
                    ignore_errors: bool = False) -> None:
    """Drop a Zarr array.

    Args:
        partition: The partition that contains the array to drop.
        base_dir: Base directory for the Zarr array.
        fs: The filesystem used to delete the Zarr array.
        variable: The name of the variable to drop.
        ignore_errors: If True, ignore errors when dropping the array.
    """
    try:
        fs.rm(join_path(partition, variable), recursive=True)
        fs.invalidate_cache(partition)
    # pylint: disable=broad-except
    # We don't want to fail on errors.
    except Exception:
        if not ignore_errors:
            raise
    # pylint: enable=broad-except


def _load_one_dataset(
    args: tuple[tuple[tuple[str, int], ...], list[slice]],
    *,
    base_dir: str,
    delayed: bool,
    fs: fsspec.AbstractFileSystem,
    selected_variables: Iterable[str] | None,
    view_ref: collection.Collection,
    variables: Sequence[str],
) -> tuple[dataset.Dataset, str] | None:
    """Load a dataset from a partition stored in the reference collection and
    merge it with the variables defined in this view.

    Args:
        args: tuple containing the partition's keys and its indexer.
        base_dir: Base directory of the view.
        delayed: If True, load the dataset lazily.
        fs: The file system used to access the variables in the view.
        selected_variables: The list of variable to retain from the view
            reference.
        view_ref: The view reference.
        variables: The variables to retain from the view

    Returns:
        The dataset and the partition's path.
    """
    partition_scheme: tuple[tuple[str, int], ...]
    slices: list[slice]

    partition_scheme, slices = args
    partition: str = view_ref.partitioning.join(partition_scheme, fs.sep)
    zds: dataset.Dataset = open_zarr_group(
        join_path(view_ref.partition_properties.dir, partition),
        view_ref.fs,
        delayed=delayed,
        selected_variables=selected_variables)
    if zds is None:
        return None

    # If the user has not selected any variables in the reference view. In this
    # case, the dataset is built from all the variables selected in the view.
    if len(zds.dimensions) == 0:
        return dataset.Dataset(
            [
                open_zarr_array(
                    zarr.open(  # type: ignore[arg-type]
                        fs.get_mapper(join_path(base_dir, partition,
                                                variable)),
                        mode='r',
                    ),
                    variable,
                    delayed=delayed) for variable in variables
            ],
            attrs=zds.attrs), partition

    # pylint: disable=expression-not-assigned
    # We use the set notation to evaluate the generator.
    {
        zds.add_variable(  # type: ignore[func-returns-value]
            item.metadata(),
            item.array,
        )
        for item in (
            open_zarr_array(
                zarr.open(  # type: ignore[arg-type]
                    fs.get_mapper(join_path(base_dir, partition, variable)),
                    mode='r',
                ),
                variable,
                delayed=delayed) for variable in variables)
    }
    # pylint: enable=expression-not-assigned

    # Apply indexing if needed.
    if len(slices):
        dim: str = view_ref.partition_properties.dim
        ds_list: list[dataset.Dataset] = [
            zds.isel({dim: indexer}) for indexer in slices
        ]
        zds = ds_list.pop(0)
        if ds_list:
            zds = zds.concat(ds_list, dim)
    return zds, partition


def _assert_variable_handled(reference: meta.Dataset, view: meta.Dataset,
                             variable: str) -> None:
    """Assert that a variable belongs to a view.

    Args:
        reference: The reference dataset.
        view: The view dataset.
        variable: The variable to check.
    """
    if variable in reference.variables:
        raise ValueError(f'Variable {variable} is read-only')
    if variable not in view.variables:
        raise ValueError(f'Variable {variable} does not exist')


def _load_datasets_list(
    *,
    client: dask.distributed.Client,
    base_dir: str,
    delayed: bool,
    fs: fsspec.AbstractFileSystem,
    view_ref: collection.Collection,
    metadata: meta.Dataset,
    partitions: Iterable[str],
    selected_variables: Iterable[str] | None = None,
) -> Iterator[tuple[dataset.Dataset, str]]:
    """Load datasets from a list of partitions.

    Args:
        client: The client used to load the datasets.
        base_dir: Base directory of the view.
        delayed: If True, load the dataset lazily.
        fs: The file system used to access the variables in the view.
        view_ref: The view reference.
        metadata: The metadata of the dataset.
        partitions: The list of partitions to load.
        selected_variables: The list of variable to retain from the view

    Returns:
        The datasets and their paths.
    """
    arguments: tuple[tuple[tuple[tuple[str, int], ...], list], ...] = tuple(
        (view_ref.partitioning.parse(item), []) for item in partitions)
    futures: list[dask.distributed.Future] = client.map(
        _load_one_dataset,
        arguments,
        base_dir=base_dir,
        delayed=delayed,
        fs=fs,
        selected_variables=view_ref.metadata.select_variables(
            keep_variables=selected_variables),
        view_ref=client.scatter(view_ref),
        variables=metadata.select_variables(selected_variables))

    return filter(lambda item: item is not None,
                  client.gather(futures))  # type: ignore[arg-type]


def _assert_have_variables(metadata: meta.Dataset) -> None:
    """Assert that the current view has variables.

    Args:
        metadata: The metadata of the dataset.
    """
    if not metadata.variables:
        raise ValueError('The view has no variables')


def _select_overlap(
    arguments: tuple[dataset.Dataset, str],
    datasets_list: tuple[tuple[dataset.Dataset, str], ...],
    depth: int,
    view_ref: collection.Collection,
) -> tuple[dataset.Dataset, slice]:
    """Select the neighboring partitions of a given partition.

    Args:
        arguments: The partition to select the neighbors of.
        datasets_list: The list of partitions.
        depth: The depth of the overlap.
        view_ref: The view reference.

    Returns:
        The neighboring partitions.
    """

    def calculate_slice(
            selected_datasets: list[tuple[dataset.Dataset, str]]) -> slice:
        """Compute the slice of the selected dataset (without overlap)."""
        start = 0
        indices = slice(0, 0, None)
        for zds, selected_partition in selected_datasets:
            size = zds.dimensions[view_ref.partition_properties.dim]
            indices = slice(start, start + size, None)
            if partition == selected_partition:
                break
            start += size
        return indices

    # The local function is not taken into account for counting
    # locals.
    partition: str = arguments[1]
    where: int = next(ix for ix, item in enumerate(datasets_list)
                      if item[1] == partition)

    # Search for the overlapping partitions
    selected_datasets: list[tuple[dataset.Dataset, str]] = [
        datasets_list[ix] for ix in range(where - depth, where + depth + 1)
        if 0 <= ix < len(datasets_list)
    ]

    # Build the dataset for the selected partitions.
    groups: list[dataset.Dataset] = [ds for ds, _ in selected_datasets]
    zds: dataset.Dataset = groups.pop(0)
    zds = zds.concat(groups, view_ref.partition_properties.dim)

    return zds, calculate_slice(selected_datasets)


def _wrap_update_func(
    func: collection.UpdateCallable,
    fs: fsspec.AbstractFileSystem,
    *args,
    **kwargs,
) -> ViewUpdateCallable:
    """Wrap an update function taking a list of partition's dataset and
    partition's path as input and returning None.

    Args:
        func: The update function.
        fs: The file system used to access the variables in the view.
        *args: The arguments of the update function.
        **kwargs: The keyword arguments of the update function.

    Returns:
        The wrapped function.
    """

    def wrap_function(parameters: Iterable[tuple[dataset.Dataset, str]],
                      base_dir: str) -> None:
        """Wrap the function to be applied to the dataset."""
        for zds, partition in parameters:
            # Applying function on partition's data
            dictionary: dict[str, ArrayLike] = func(zds, *args, **kwargs)
            tuple(
                update_zarr_array(  # type: ignore[func-returns-value]
                    dirname=join_path(base_dir, partition, varname),
                    array=array,
                    fs=fs,
                ) for varname, array in dictionary.items())

    return wrap_function


def _wrap_update_func_overlap(
    datasets_list: tuple[tuple[dataset.Dataset, str], ...],
    depth: int,
    func: collection.UpdateCallable,
    fs: fsspec.AbstractFileSystem,
    view_ref: collection.Collection,
    trim: bool,
    *args,
    **kwargs,
) -> ViewUpdateCallable:
    """Wrap an update function taking a list of partition's dataset and
    partition's path as input and returning None.

    Args:
        datasets_list: The list of datasets and their paths.
        depth: The depth of the overlap.
        func: The update function.
        fs: The file system used to access the variables in the view.
        view_ref: The view reference.
        trim: If True, trim the dataset to the overlap.
        *args: The arguments of the update function.
        **kwargs: The keyword arguments of the update function.

    Returns:
        The wrapped function.
    """
    dim = view_ref.partition_properties.dim

    if depth < 0:
        raise ValueError('The depth must be positive')

    def wrap_function(parameters: Iterable[tuple[dataset.Dataset, str]],
                      base_dir: str) -> None:
        """Wrap the function to be applied to the dataset."""
        zds: dataset.Dataset
        indices: slice

        for zds, partition in parameters:
            zds, indices = _select_overlap((zds, partition), datasets_list,
                                           depth, view_ref)
            # pylint: disable=duplicate-code
            # False positive with the function _wrap_update_func_with_overlap
            # defined in the module zcollection.collection.detail
            _update_with_overlap(*args,
                                 func=func,
                                 zds=zds,
                                 indices=indices,
                                 dim=dim,
                                 fs=fs,
                                 path=join_path(base_dir, partition),
                                 trim=trim,
                                 **kwargs)
            # pylint: enable=duplicate-code

    return wrap_function


def _calculate_axis_reference(
        path: str,
        view_ref: collection.Collection,
        group: zarr.Group | None = None) -> AxisReference:
    """Compute the axis reference of a partition (checksum, array and dimension
    name).

    Args:
        path: The path of the partition.
        fs: The file system used to access the variables in the view.
        view_ref: The view reference.
        group: The group to compute the checksum of (if None, the group is
            reload from the file system).

    Returns:
        The axis reference of the partition.
    """
    store: zarr.Group
    store = group or zarr.open_consolidated(  # type: ignore[arg-type]
        view_ref.fs.get_mapper(path), )
    array: zarr.Array = store[view_ref.axis]  # type: ignore[arg-type]
    axis: NDArray = array[...]
    checksum: str = hashlib.sha256(axis.tobytes()).hexdigest()
    return AxisReference(axis, array.attrs[DIMENSIONS][0], checksum)


def _write_checksum_array(
    partition: str,
    fs: fsspec.AbstractFileSystem,
    axis_ref: AxisReference,
) -> None:
    """Write the checksum of a partition to a file.

    Args:
        partition: The path of the partition.
        fs: The file system used to access the variables in the view.
        axis_ref: The axis reference of the partition.
    """
    array: zarr.Array
    checksum_path: str = join_path(partition, CHECKSUM_FILE)
    mapper: fsspec.FSMap = fs.get_mapper(checksum_path)
    if fs.exists(checksum_path):
        array = zarr.open(mapper)  # type: ignore[arg-type]
    else:
        array = zarr.create(shape=axis_ref.array.shape,
                            dtype=axis_ref.array.dtype,
                            store=mapper)
        array.attrs[DIMENSIONS] = axis_ref.dimension
    array[...] = axis_ref.array
    array.attrs[CHECKSUM_ATTR] = axis_ref.checksum
    fs.invalidate_cache(checksum_path)


def _load_checksum_array(
    partition: str,
    fs: fsspec.AbstractFileSystem,
) -> zarr.Array:
    """Load the checksum of a partition from a file.

    Args:
        partition: The path of the partition.
        fs: The file system used to access the variables in the view.

    Returns:
        The checksum of the partition and the axis of the reference partition.
    """
    checksum_path: str = join_path(partition, CHECKSUM_FILE)
    mapper: fsspec.FSMap = fs.get_mapper(checksum_path)
    return zarr.open_array(mapper)


def _write_checksum(
    partition: str,
    base_dir: str,
    view_ref: collection.Collection,
    fs: fsspec.AbstractFileSystem,
    group: zarr.Group | None = None,
) -> None:
    """Write the checksum of a partition to a file.

    Args:
        partition: The path of the partition.
        base_dir: The base directory of the view.
        view_ref: The view reference.
        fs: The file system used to access the variables in the view.
        group: The group to compute the checksum of (if None, the group is
            reload from the file system).
    """
    partition_ref: str = join_path(
        view_ref.partition_properties.dir,
        str(pathlib.Path(partition).relative_to(base_dir).as_posix()))
    _write_checksum_array(
        partition, fs,
        _calculate_axis_reference(partition_ref, view_ref, group=group))


def _sync_partition(
    metadata: meta.Dataset,
    partition: str,
    base_dir: str,
    fs: fsspec.AbstractFileSystem,
    view_ref: collection.Collection,
) -> None:
    """Sync a partition: create the partition and the underlying variables.

    Args:
        metadata: The dataset to sync.
        partition: The partition to sync.
        base_dir: The base directory of the view.
        fs: The file system used to access the variables in the view.
        view_ref: The view reference.
    """
    mapper: fsspec.FSMap = view_ref.fs.get_mapper(
        join_path(view_ref.partition_properties.dir, partition))
    group: zarr.Group = zarr.open_consolidated(  # type: ignore
        mapper, mode='r')
    path: str = join_path(base_dir, partition)
    try:
        for variable in metadata.variables.values():
            template: meta.Variable = \
                view_ref.metadata.search_same_dimensions_as(variable)
            _create_zarr_array((partition, group),
                               base_dir=base_dir,
                               fs=fs,
                               template=template.name,
                               variable=variable,
                               invalidate_cache=False)
        _write_checksum(path, base_dir, view_ref, fs, group)
        fs.invalidate_cache(path)

    except Exception as exc:
        fs.rm(path, recursive=True)
        fs.invalidate_cache(path)
        raise exc from None


def _extend_partition(
    partition: str,
    fs: fsspec.AbstractFileSystem,
    axis_ref: AxisReference,
) -> None:
    """Sync a partition: create the partition and the underlying variables.

    Args:
        partition: The partition to sync.
        fs: The file system used to access the variables in the view.
        axis_ref: The axis reference of the partition.
    """
    axis_name: str = axis_ref.dimension
    new_size: int = axis_ref.array.shape[0]
    try:
        for variable in fs.listdir(partition):
            array: zarr.Array = zarr.open_array(fs.get_mapper(
                variable['name']))
            dimensions: Sequence[str] = array.attrs[DIMENSIONS]
            if axis_name in dimensions:
                axis: int = dimensions.index(axis_name)
                shape = list(array.shape)
                shape[axis] = new_size
                array.resize(shape)
        _write_checksum_array(partition, fs, axis_ref)
        # fs.invalidate_cache(partition) is not done by
        # _write_checksum_array
    except Exception as exc:
        fs.rm(partition, recursive=True)
        fs.invalidate_cache(partition)
        raise exc from None


def _sync(
    partition: str,
    *,
    base_dir: str,
    fs: fsspec.AbstractFileSystem,
    view_ref: collection.Collection,
    metadata: meta.Dataset,
    dry_run: bool = False,
) -> str | None:
    """Sync the partitions of a view.

    Args:
        partition: The partition to sync.
        base_dir: The base directory of the view.
        fs: The file system used to access the variables in the view.
        view_ref: The view reference.
        metadata: The dataset to sync.
        dry_run: If True, the partition is not synced.

    Returns:
        The partition synced or None if the partition is already synced.
    """
    partition_view: str = join_path(base_dir, partition)
    if not fs.exists(partition_view):
        # The partition does not exist, so we create it.
        if not dry_run:
            _sync_partition(metadata, partition, base_dir, fs, view_ref)
        return partition

    # The partition exists, so we check if it is synced.
    partition_ref: str = join_path(view_ref.partition_properties.dir,
                                   partition)
    array: zarr.Array = _load_checksum_array(partition_view, fs)
    axis_ref: AxisReference = _calculate_axis_reference(
        partition_ref, view_ref)
    # If the checksums are different, the partition is not synced. So we
    # remove it (information between the partition and the reference are
    # not consistent)
    if axis_ref.checksum != array.attrs[CHECKSUM_ATTR]:
        # If the checksums are different, the partition is not synced. we load
        # the axis of partition's view to check if the partition is a subset of
        # the reference partition.
        axis_view: NDArray = array[:]
        if axis_ref.array.size > axis_view.size and numpy.all(
                axis_view == axis_ref.array[:axis_view.size]):
            # The view partition is a subset of the reference partition.
            # So we extend the view partition to the reference partition.
            # fs_invalid_cache is done by _extend_partition
            if not dry_run:
                _extend_partition(partition_view, fs, axis_ref)
            return partition
        # The partition is not synced, so we remove it.
        if not dry_run:
            fs.rm(partition_view, recursive=True)
            fs.invalidate_cache(partition_view)
            _sync_partition(metadata, partition, base_dir, fs, view_ref)
        return partition
    return None


def _serialize_filters(filters: collection.PartitionFilter, ) -> str:
    """Serialize a partition filter.

    Args:
        filters: The partition filter to serialize.

    Returns:
        The serialized partition filter.
    """
    stream = io.BytesIO()
    try:
        pickle.dump(filters, stream)
    except (pickle.PicklingError, AttributeError):
        warnings.warn(
            'The partition filter cannot be serialized, it will be ignored.',
            UserWarning,
            stacklevel=2)
        pickle.dump(None, stream)
    return base64.b64encode(stream.getvalue()).decode('utf-8')


def _deserialize_filters(filters: str) -> collection.PartitionFilter:
    """Unserialize a partition filter.

    Args:
        filters: The partition filter to unserialize.

    Returns:
        The unserialized partition filter.
    """
    stream = io.BytesIO(base64.b64decode(filters.encode('utf-8')))
    try:
        return pickle.load(stream)
    except pickle.UnpicklingError:
        return None



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing the sync module.
========================
"""
from .. import sync


def test_no_sync():
    """Test the no_sync class."""
    touch = False
    with sync.NoSync() as _:
        touch = True
    assert touch



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Handles the partition selection expressions
===========================================
"""
from __future__ import annotations

from typing import Any, ClassVar
import ast
import dataclasses
import types


@dataclasses.dataclass
class Expression:
    """Partitioning expressions.

    Args:
        expression: The expression to be evaluated
    Raises:
        NameError: If a variable is not defined.
    Example:
        >>> expr = Expression("year==2000 and month==1 and day in range(1, 12)")
    """
    #: Compiled expression to be evaluated
    code: types.CodeType

    #: Known data members
    __slots__: tuple[str, ...] = ('code', )

    #: The builtins that are allowed in the expression.
    BUILTINS: ClassVar[dict[str, Any]] = {'range': range}

    def __init__(self, expression: str) -> None:
        self.code = compile(ast.parse(expression, mode='eval'), ' ', 'eval')

    def __call__(self, variables: dict[str, Any]) -> Any:
        try:
            __locals: dict[str, Any] = {
                name: variables[name]
                for name in self.code.co_names if name not in self.BUILTINS
            }
            # pylint: disable=eval-used
            # The eval function is used here to evaluate a simple expression.
            # The only builtin functions allowed is the range function.
            return eval(self.code, {'__builtins__': self.BUILTINS}, __locals)
            # pylint: enable=eval-used
        except KeyError as err:
            raise NameError(f'name {err!s} is not defined') from err
